# Copyright 2018-2026 Jérôme Dumonteil
# Authors (odfdo project): jerome.dumonteil@gmail.com
from __future__ import annotations

import subprocess
import sys
from pathlib import Path

import pytest

from odfdo.document import Document
from odfdo.scripts import meta_print
from odfdo.scripts.meta_print import main as main_script
from odfdo.scripts.meta_print import main_meta_print, parse_cli_args

SCRIPT = Path(meta_print.__file__)
SOURCE = "user_fields.odt"


def run_params(params: list):
    command = [sys.executable, SCRIPT] + params
    proc = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = proc.communicate()
    return out, err, proc.returncode


def test_meta_print_no_param():
    params = []
    out, _err, exitcode = run_params(params)
    assert exitcode == 1
    assert b"usage:" in out
    assert b"Error:" in out


# direct access to internal function


def test_meta_print_2_no_param_on_main_function(monkeypatch, capsys):
    with pytest.raises(SystemExit) as result:
        monkeypatch.setattr(sys, "argv", [])
        main_script()
        assert result.type is SystemExit
        assert result.value.code >= 1
    captured = capsys.readouterr()

    assert "usage" in captured.out


def test_meta_print_2_no_param(capsys):
    params = parse_cli_args([])

    with pytest.raises(SystemExit) as result:
        main_meta_print(params)
        assert result.type is SystemExit
        assert result.value.code == 1
    captured = capsys.readouterr()

    assert "usage" in captured.out


def test_meta_print_2_no_file(capsys):
    params = parse_cli_args(["-i", "none_file1", "-o", "none_file2"])

    with pytest.raises(SystemExit) as result:
        main_meta_print(params)
        assert result.type is SystemExit
        assert result.value.code == 1
    captured = capsys.readouterr()

    assert "usage" in captured.out
    assert "FileNotFoundError" in captured.out


def test_meta_2_print_language(capsys, samples):
    source = str(samples(SOURCE))
    params = parse_cli_args(["-l", "-i", source])

    main_meta_print(params)
    captured = capsys.readouterr()

    assert "Default style language: fr-FR" in captured.out
    assert "Statistic:" in captured.out


def test_meta_print_2_odf_version(capsys, samples):
    source = str(samples(SOURCE))
    params = parse_cli_args(["-v", "-i", source])

    main_meta_print(params)
    captured = capsys.readouterr()

    assert "OpenDocument format version: 1.3" in captured.out
    assert "Statistic:" in captured.out


def test_meta_print_2_text(capsys, samples):
    source = str(samples(SOURCE))
    params = parse_cli_args(["-i", source])

    main_meta_print(params)
    captured = capsys.readouterr()

    assert "Creation date:" in captured.out
    assert "Statistic:" in captured.out
    assert "Object count:" in captured.out


def test_meta_print_2_json(capsys, samples):
    source = str(samples(SOURCE))
    params = parse_cli_args(["-j", "-i", source])

    main_meta_print(params)
    captured = capsys.readouterr()

    assert "meta:creation-date" in captured.out
    assert "dc:date" in captured.out
    assert "meta:table-count" in captured.out
    assert "meta:generator" in captured.out
    assert "meta:initial-creator" in captured.out


def test_meta_print_2_json_save(tmp_path, samples):
    source = str(samples(SOURCE))
    dest = tmp_path / "json_data.json"
    params = parse_cli_args(["-j", "-i", source, "-o", str(dest)])

    main_meta_print(params)

    result = dest.read_text(encoding="utf8")
    assert "meta:creation-date" in result
    assert "dc:date" in result
    assert "meta:table-count" in result
    assert "meta:generator" in result
    assert "meta:initial-creator" in result


def test_meta_print_2_text_from_stdin(monkeypatch, capsys, samples):
    mock_document = Document(samples(SOURCE))

    def mock_read_document(input_path: str | None):
        assert input_path is None
        return mock_document

    monkeypatch.setattr("odfdo.scripts.meta_print.read_document", mock_read_document)

    params = parse_cli_args([])

    main_meta_print(params)
    captured = capsys.readouterr()

    assert "Creation date:" in captured.out
    assert "Statistic:" in captured.out
    assert "Object count:" in captured.out



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.form_controls import FormListbox


@pytest.fixture
def form_listbox() -> Iterable[FormListbox]:
    yield FormListbox(
        name="some name",
        control_implementation="some implem",
        title="some title",
        data_field="",
        linked_cell="",
        disabled=False,
        dropdown=True,
        printable=True,
        bound_column="bound",
        list_linkage_type="selection",
        multiple=True,
        list_source="some source",
        list_source_type="table",
        source_cell_range="range",
        size=7,
        tab_index=4,
        tab_stop=False,
        xml_id="control1",
    )


def test_form_listbox_class():
    form = FormListbox()
    assert isinstance(form, FormListbox)


def test_form_listbox_minimal_tag():
    form = Element.from_tag("<form:listbox/>")
    assert isinstance(form, FormListbox)


def test_form_listbox_serialize():
    form = FormListbox()
    assert form.serialize() == "<form:listbox/>"


def test_form_listbox_repr():
    form = FormListbox()
    assert repr(form) == "<FormListbox name=None xml_id=None>"


def test_form_listbox_name(form_listbox):
    assert form_listbox.name == "some name"


def test_form_listbox_control_implementation(form_listbox):
    assert form_listbox.control_implementation == "some implem"


def test_form_listbox_disabled(form_listbox):
    assert form_listbox.disabled is False


def test_form_listbox_printable(form_listbox):
    assert form_listbox.printable is True


def test_form_listbox_tab_index(form_listbox):
    assert form_listbox.tab_index == 4


def test_form_listbox_tab_stop(form_listbox):
    assert form_listbox.tab_stop is False


def test_form_listbox_xml_id(form_listbox):
    assert form_listbox.xml_id == "control1"


def test_form_listbox_form_id(form_listbox):
    assert form_listbox.form_id == "control1"


def test_form_listbox_title(form_listbox):
    assert form_listbox.title == "some title"


def test_form_listbox_dropdown(form_listbox):
    assert form_listbox.dropdown is True


def test_form_listbox_bound_column(form_listbox):
    assert form_listbox.bound_column == "bound"


def test_form_listbox_list_linkage_type_1(form_listbox):
    assert form_listbox.list_linkage_type == "selection"


def test_form_listbox_list_linkage_type_2(form_listbox):
    with pytest.raises(ValueError):
        form_listbox.list_linkage_type = "bad"


def test_form_listbox_list_linkage_type_3(form_listbox):
    form_listbox.list_linkage_type = None
    assert form_listbox.list_linkage_type is None


def test_form_listbox_list_linkage_type_4(form_listbox):
    form_listbox.list_linkage_type = "selection-indices"
    assert form_listbox.list_linkage_type == "selection-indices"


def test_form_listbox_multiple(form_listbox):
    assert form_listbox.multiple is True


def test_form_listbox_list_source(form_listbox):
    assert form_listbox.list_source == "some source"


def test_form_listbox_list_source_type_1(form_listbox):
    assert form_listbox.list_source_type == "table"


def test_form_listbox_list_source_type_2(form_listbox):
    with pytest.raises(ValueError):
        form_listbox.list_source_type = "bad"


def test_form_listbox_list_source_type_3(form_listbox):
    form_listbox.list_source_type = None
    assert form_listbox.list_source_type is None


def test_form_listbox_list_source_type_4(form_listbox):
    form_listbox.list_source_type = "sql"
    assert form_listbox.list_source_type == "sql"


def test_form_listbox_source_cell_range(form_listbox):
    assert form_listbox.source_cell_range == "range"


def test_form_listbox_xforms_bind():
    form = FormListbox(
        name="some name",
        control_implementation="some implem",
        xml_id="control1",
        xforms_bind="bind id",
    )
    assert form.xforms_bind == "bind id"


def test_form_listbox_xforms_list_source():
    form = FormListbox(
        name="some name",
        control_implementation="some implem",
        xml_id="control1",
        xforms_list_source="source",
    )
    assert form.xforms_list_source == "source"


def test_form_listbox_form_id_2():
    form = FormListbox(
        name="some name",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
    )
    assert form.form_id == "control1"


def test_form_listbox_data_field(form_listbox):
    assert form_listbox.data_field == ""


def test_form_listbox_linked_cell(form_listbox):
    assert form_listbox.linked_cell == ""


def test_form_listbox_size_1(form_listbox):
    assert form_listbox.size == 7


def test_form_listbox_size_2():
    form = FormListbox(
        name="some name",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
    )
    assert form.size is None


def test_form_listbox_size_3():
    form = FormListbox(
        name="some name",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
        size=25,
    )
    form.size = None
    assert form.size is None


def test_form_listbox_size_4():
    form = FormListbox(
        name="some name",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
    )
    form.size = -4
    assert form.size == 0



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>

import pytest

from odfdo.style import Style


def test_table_cell_border():
    style = Style("table-cell", border="0.002cm")
    assert style.serialize() == (
        "<style:style "
        'style:family="table-cell"><style:table-cell-properties '
        'fo:border="0.002cm"/></style:style>'
    )


def test_table_cell_border_border_left():
    style = Style("table-cell", border="0.002cm", border_left="0.002cm")
    assert style.serialize() == (
        "<style:style "
        'style:family="table-cell"><style:table-cell-properties '
        'fo:border="0.002cm"/></style:style>'
    )


def test_table_cell_border_top():
    style = Style("table-cell", border_top="0.002cm")
    assert style.serialize() in (
        (
            "<style:style "
            'style:family="table-cell"><style:table-cell-properties '
            'fo:border-bottom="none" fo:border-left="none" '
            'fo:border-right="none" fo:border-top="0.002cm"/>'
            "</style:style>"
        ),
        (
            "<style:style "
            'style:family="table-cell"><style:table-cell-properties '
            'fo:border-top="0.002cm" fo:border-right="none" '
            'fo:border-bottom="none" fo:border-left="none"/>'
            "</style:style>"
        ),
        (
            "<style:style "
            'style:family="table-cell"><style:table-cell-properties '
            'fo:border-top="0.002cm" fo:border-left="none" '
            'fo:border-right="none" fo:border-bottom="none"/>'
            "</style:style>"
        ),
    )


def test_table_cell_border_all():
    style = Style(
        "table-cell",
        border_top="0.001cm",
        border_right="0.002cm",
        border_bottom="0.003cm",
        border_left="0.004cm",
    )

    assert style.serialize() in (
        (
            "<style:style "
            'style:family="table-cell"><style:table-cell-properties '
            'fo:border-top="0.001cm" fo:border-right="0.002cm" '
            'fo:border-bottom="0.003cm" fo:border-left="0.004cm"/>'
            "</style:style>"
        ),
        (
            "<style:style "
            'style:family="table-cell"><style:table-cell-properties '
            'fo:border-bottom="0.003cm" fo:border-left="0.004cm" '
            'fo:border-right="0.002cm" fo:border-top="0.001cm"/>'
            "</style:style>"
        ),
    )


def test_table_cell_shadow():
    style = Style("table-cell", shadow="#808080 0.176cm 0.176cm")
    assert style.serialize() == (
        "<style:style "
        'style:family="table-cell"><style:table-cell-properties '
        'style:shadow="#808080 0.176cm 0.176cm"/></style:style>'
    )


def test_table_row_height():
    style = Style("table-row", height="5cm")
    assert style.serialize() == (
        "<style:style "
        'style:family="table-row"><style:table-row-properties '
        'style:row-height="5cm"/></style:style>'
    )


def test_table_column_width():
    style = Style("table-column", width="5cm")
    assert style.serialize() == (
        "<style:style "
        'style:family="table-column"><style:table-column-properties '
        'style:column-width="5cm"/></style:style>'
    )


def test_table_width():
    style = Style("table", "tab1", width="5cm")
    assert style.serialize() == (
        "<style:style "
        'style:family="table" '
        'style:name="tab1">'
        "<style:table-properties "
        'style:width="5cm"/>'
        "</style:style>"
    )


def test_table_width_align_left():
    style = Style("table", "tab1", width="5cm", align="left")
    assert style.serialize() == (
        "<style:style "
        'style:family="table" '
        'style:name="tab1">'
        "<style:table-properties "
        'style:width="5cm" '
        'table:align="left"/>'
        "</style:style>"
    )


def test_table_width_align_left_marge():
    style = Style("table", "tab1", width="5cm", align="left", margin_left="2cm")
    assert style.serialize() == (
        "<style:style "
        'style:family="table" '
        'style:name="tab1">'
        "<style:table-properties "
        'fo:margin-left="2cm" '
        'style:width="5cm" '
        'table:align="left"/>'
        "</style:style>"
    )


def test_table_align_wrong():
    with pytest.raises(ValueError):
        Style("table", "tab1", align="wrong")



#!/usr/bin/env python
"""Create a basic document containing some paragraphs and headers, add some
annotations. Annotations are notes that don't appear in the document but
typically on a side bar in a desktop application. So they are not printed.
"""

import os
from pathlib import Path

from odfdo import Document, Header, Paragraph

_DOC_SEQUENCE = 40
DATA = Path(__file__).parent / "data"
LOREM = (DATA / "lorem.txt").read_text(encoding="utf8")
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "basic_annotations"
TARGET = "document.odt"


def save_new(document: Document, name: str) -> None:
    """Save a recipe result Document."""
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    new_path = OUTPUT_DIR / name
    print("Saving:", new_path)
    document.save(new_path, pretty=True)


def make_document() -> Document:
    """Generate a basic document containing some paragraphs and headers."""
    document = Document("text")
    body = document.body
    body.clear()

    level_1_title = Header(1, "Main title")
    body.append(level_1_title)

    for idx in range(3):
        level_2_title = Header(2, f"title {idx}")
        body.append(level_2_title)
        paragraph = Paragraph(LOREM[:240])
        body.append(paragraph)

    return document


def make_annotations(document: Document) -> None:
    """Add some annotation on each pragraph of the document."""
    word_position = 0
    for paragraph in document.body.paragraphs:
        # choosing some word of the paragraph to insert the note
        word_position += 3
        some_word = str(paragraph).split()[word_position]

        # Adding Annotation
        paragraph.insert_annotation(
            # The word after what the annotation is inserted:
            after=some_word,
            # The annotation itself, at the end of the page:
            body=f"It's so easy ! (after {some_word!r})",
            # The author of the annotation:
            creator="Bob",
            # A datetime value, by default datetime.now():
            # date= xxx
        )


def main():
    document = make_document()
    make_annotations(document)
    test_unit(document)
    save_new(document, TARGET)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    assert len(document.body.get_annotations(creator="Bob")) == 3


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Mixin classes for Markdown methods."""

from __future__ import annotations

import re
from collections.abc import Callable
from copy import deepcopy
from itertools import chain
from typing import Any, NamedTuple

MD_GLOBAL = {}

RE_STAR6 = re.compile(r"(?<!\\)(\*{6})")
RE_STAR4 = re.compile(r"(?<!\\)(\*{4})")
RE_UND2 = re.compile(r"(?<!\\)(_{2})")


class LIStyle(NamedTuple):
    name: str
    format: str


class SplitSpace(NamedTuple):
    start: str
    word: str
    end: str


def _set_global(doc: Any) -> None:
    MD_GLOBAL["document"] = doc
    MD_GLOBAL["list_level"] = {}
    MD_GLOBAL["footnote"] = []
    MD_GLOBAL["endnote"] = []


def _copy_global() -> dict[str, Any]:
    return deepcopy(MD_GLOBAL)


def _restore_global(data: dict[str, Any]) -> None:
    for key, val in data.items():
        MD_GLOBAL[key] = val


def _get_list_counter(name: str, level: int) -> int:
    ref = f"{name}_{level}"
    list_level = MD_GLOBAL.get("list_level", {})
    last_level = list_level.get("last_level", 0)
    if level > last_level:
        last = 0
    else:
        last = list_level.get(ref, 0)
    list_level["last_level"] = level
    counter = last + 1
    list_level[ref] = counter
    return counter


def _release_list_counter(level: int) -> None:
    list_level = MD_GLOBAL.get("list_level", {})
    list_level["last_level"] = level


def _strip_left_spaces(text: str) -> str:
    return RE_STAR4.sub("", RE_STAR6.sub("", RE_UND2.sub("", text.lstrip(" "))))


def _md_swap_spaces(word: str) -> SplitSpace:
    if not word:
        return SplitSpace("", "", "")
    space_before = 0
    while word.startswith(" "):
        space_before += 1
        word = word[1:]
    space_after = 0
    while word.endswith(" "):
        space_after += 1
        word = word[:-1]
    return SplitSpace(" " * space_before, word, " " * space_after)


def _md_escape(text: str | None) -> str:
    if not text:
        return ""
    return (
        # non break space is no understood as char
        text.replace(" ", r" ")  # noqa: RUF001
        .replace("#", r"\#")
        .replace(r"\*", "*")
        .replace("*", r"\*")
        .replace(r"\_", r"_")
        .replace("_", r"\_")
        .replace("-", r"\-")
        .replace(r"\`", "`")
        .replace("`", r"\`")
        .replace(r"\~", "~")
        .replace("~", r"\~")
        .replace("|", r"\|")
    )


def _as_italic(text: str | None) -> str:
    text = _md_escape(text)
    if not text.strip():
        return text
    word = _md_swap_spaces(text)
    return f"{word.start}_{word.word}_{word.end}"


def _as_bold(text: str | None) -> str:
    text = _md_escape(text)
    if not text.strip():
        return text
    word = _md_swap_spaces(text)
    return f"{word.start}**{word.word}**{word.end}"


def _as_bold_italic(text: str | None) -> str:
    text = _md_escape(text)
    if not text.strip():
        return text
    word = _md_swap_spaces(text)
    return f"{word.start}***{word.word}***{word.end}"


def _as_fixed(text: str | None) -> str:
    text = _md_escape(text)
    if not text.strip():
        return text
    return f"`{text}`"


def _as_strike(text: str | None) -> str:
    text = _md_escape(text)
    if not text.strip():
        return text
    return f"~~{text}~~"


def _as_none(text: str | None) -> str:
    return _md_escape(text)


class MDStyle:
    def _md_is_fixed_paragraph(self) -> bool:
        if self.tag != "text:p" or not self.style:
            return False
        document = MD_GLOBAL.get("document")
        if not document:
            return False
        style = document.get_style("paragraph", self.style)
        if not style:
            return False
        parent_style = document.get_parent_style(style)
        if parent_style:
            prop = parent_style.get_text_properties()
            prop.update({k: v for k, v in style.get_text_properties().items() if v})
        else:
            prop = style.get_text_properties()
        return bool(prop["fixed"])

    def _md_styling(self) -> Callable:
        def get_text_props(document: Any, name: str) -> dict[str, Any]:
            prop: dict[str, Any] = {}
            style = document.get_style("text", name)
            if not style:
                style = document.get_style("paragraph", name)
            if not style:
                return prop
            parent_style = document.get_parent_style(style)
            if parent_style:
                prop = parent_style.get_text_properties()
                prop.update({k: v for k, v in style.get_text_properties().items() if v})
            else:
                prop = style.get_text_properties()
            return prop

        if not self.style:
            return _as_none
        document = MD_GLOBAL.get("document")
        if not document:
            return _as_none
        prop = get_text_props(document, self.style)
        if not prop:
            return _as_none
        if prop["italic"]:
            if prop["bold"]:
                return _as_bold_italic
            else:
                return _as_italic
        elif prop["bold"]:
            return _as_bold
        elif prop["fixed"]:
            return _as_fixed
        elif prop["strike"]:
            return _as_strike
        return _as_none


class MDDocument:
    def _md_collect(self) -> list[str]:
        return [
            item
            for item in chain.from_iterable(
                child._md_collect() for child in self.body.children
            )
            if item
        ]

    def _markdown_export(self) -> str:
        def join_fixed_lines(items: list[str]) -> list[str]:
            joined = []
            previous = ""
            for item in items:
                if item.startswith("```\n") and previous.endswith("```\n"):
                    content = previous[:-4] + item[4:]
                    previous = content
                else:
                    joined.append(previous)
                    previous = item
            joined.append(previous)
            return joined

        _set_global(self)
        md_list = self._md_collect()
        joined: list[str] = join_fixed_lines(md_list)
        if MD_GLOBAL["footnote"]:
            joined.extend(MD_GLOBAL["footnote"])
            joined[-1] += "\n"
        if MD_GLOBAL["endnote"]:
            joined.extend(MD_GLOBAL["endnote"])
        raw_text = "\n".join(x for x in joined if x.strip())
        _set_global(None)
        return "\n".join(x.rstrip(" ") for x in raw_text.split("\n"))


class MDBase(MDStyle):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        return _as_none(self.inner_text) + post_styler(self.tail)  # type: ignore

    def _md_collect(self) -> list[str]:
        return list(chain.from_iterable(child._md_collect() for child in self.children))


class MDToc(MDBase):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        index_body = self.get_element("text:index-body")
        if index_body is None:
            return ""
        result = []
        for element in index_body.children:
            if element.tag == "text:index-title":
                result.append(_as_bold(element.inner_text.strip()))
                continue
            result.append(element._md_format())
        return "\n\n".join(x for x in result if x)

    def _md_collect(self) -> list[str]:
        if content := self._md_format():
            return [content + "\n"]
        return []


class MDNote(MDBase):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        citation = f"[{self.citation}]"
        if self.note_class == "footnote":
            MD_GLOBAL["footnote"].append(str(self))
        else:
            MD_GLOBAL["endnote"].append(str(self))
        return citation + str(post_styler(self.tail))

    def _md_collect(self) -> list[str]:
        return [self._md_format() + "\n"]


class MDTail(MDStyle):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        return post_styler(self.tail)  # type: ignore


class MDZap(MDStyle):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        return ""

    def _md_collect(self) -> list[str]:
        return []


class MDSpacer(MDStyle):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        return self.text + post_styler(self.tail)  # type: ignore


class MDTab(MDStyle):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        return "    " + post_styler(self.tail)  # type: ignore


class MDLineBreak(MDStyle):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        return "\\\n" + str(post_styler(self.tail))

    def _md_collect(self) -> list[str]:
        return [self._md_format()]


class MDParagraph(MDStyle):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        styler = self._md_styling()
        acc = [styler(self.text)]
        acc.extend([child._md_format(styler) for child in self.children])
        acc.append(post_styler(self.tail))
        return _strip_left_spaces("".join(x for x in acc if x))

    def _md_collect_fixed_text(self) -> str:
        acc = ["```\n", self.inner_text, "\n```"]
        if tail := _as_none(self.tail):
            acc.append("\n")
            acc.append(tail)
        content = "".join(x for x in acc if x)
        return content

    def _md_collect_list_item_style(self) -> LIStyle:
        if not self.style:
            return LIStyle("", "")
        document = MD_GLOBAL.get("document")
        if not document:
            return LIStyle("", "")
        style = document.get_style("paragraph", self.style)
        if not style:
            return LIStyle("", "")
        list_style = document.get_list_style(style)
        if not list_style:
            return LIStyle("", "")
        level_style_number = list_style.get_element("text:list-level-style-number")
        if not level_style_number:
            return LIStyle("", "")
        num_format = level_style_number.style_num_format or ""
        return LIStyle(list_style.name, num_format)

    def _md_collect(self) -> list[str]:
        if self._md_is_fixed_paragraph():
            content = self._md_collect_fixed_text()
        else:
            content = self._md_format()
        if content:
            return [content + "\n"]
        return []


class MDHeader(MDParagraph):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        acc = [_as_none(self.text)]
        acc.extend([child._md_format() for child in self.children])
        acc.append(post_styler(self.tail))
        content = _strip_left_spaces("".join(x for x in acc if x))
        level = self.get_attribute_integer("text:outline-level") or 0
        if not level:
            return content
        level = min(level, 6)
        return f"{'#' * level} {content}"


class MDListItem(MDParagraph):
    def _md_list_marker(
        self,
        level: int = 0,
        li_style: LIStyle | None = None,
    ) -> str:
        if li_style is None or not li_style.format:
            return "   " * level + " -  "
        counter = _get_list_counter(li_style.name, level)
        return "   " * level + f" {counter}. "

    def _md_format(self, post_styler: Callable = _as_none, level: int = 0) -> str:
        acc = []
        for child in self.children:
            if child.tag == "text:list":
                acc.append(child._md_format(level=level + 1))
                continue
            if child.tag == "text:p":
                li_style: LIStyle = child._md_collect_list_item_style()
                acc.append(self._md_list_marker(level, li_style) + child._md_format())
            else:
                acc.append(self._md_list_marker(level) + child._md_format())
        acc.append(post_styler(self.tail))
        content = "\n".join(x for x in acc if x)
        return content

    def _md_initialize_level(self) -> None:
        _release_list_counter(0)


class MDList(MDStyle):
    def _md_format(self, post_styler: Callable = _as_none, level: int = 0) -> str:
        acc = []
        for child in self.children:
            if child.tag != "text:list-item":
                continue
            acc.append(child._md_format(level=level))
        acc.append(post_styler(self.tail))
        _release_list_counter(level + 1)
        content = "\n".join(x for x in acc if x)
        return content

    def _md_collect(self) -> list[str]:
        if content := self._md_format():
            return [content + "\n"]
        return []


class MDSpan(MDStyle):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        # acc = [self.text]
        styler = self._md_styling()
        acc = []
        if self.text:
            if styler == _as_none:
                acc.append(post_styler(self.text))
            else:
                acc.append(styler(self.text))
        acc.extend([child._md_format(styler) for child in self.children])
        acc.append(post_styler(self.tail))
        return "".join(x for x in acc if x)

    def _md_collect(self) -> list[str]:
        if content := self._md_format():
            return [content]
        return []


class MDLink(MDStyle):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        text = self.inner_text.strip()
        url = self.url
        if url and url.startswith("#"):
            url = "#"
        if text:
            svalue = f"[{text}]({url})"
        else:
            svalue = f"({url})"

        acc = [svalue]
        acc.append(post_styler(self.tail))
        content = "".join(x for x in acc if x)
        return content

    def _md_collect(self) -> list[str]:
        return [self._md_format()]


class MDDrawTextBox:
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        acc = [child._md_format() for child in self.children]
        acc.append(post_styler(self.tail))
        content = "".join(x for x in acc if x)
        return content

    def _md_collect(self) -> list[str]:
        if content := self._md_format():
            return [content]
        return []


class MDDrawFrame(MDStyle):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        url = ""
        acc = []
        if img := self.get_image():
            url = img.url
        if url:
            alter = self.svg_title or "image"
            svalue = f"![{alter}]({url})\n"
            acc.append(svalue)
        acc.extend(
            [child._md_format() for child in self.children if child.tag != "svg:title"]
        )
        acc.append(post_styler(self.tail))
        content = "".join(x for x in acc if x)
        return content

    def _md_collect(self) -> list[str]:
        if content := self._md_format():
            return [content]
        return []


class MDTable(MDStyle):
    def _md_format(self, post_styler: Callable = _as_none) -> str:
        def bars(values: list[str]) -> str:
            items = [""] + values + [""]  # noqa: RUF005
            return "|".join(items)

        def format_cell(val: Any, filler: str = " ") -> str:
            if isinstance(val, list):
                result = []
                for element in val:  # paragraph
                    styler = element._md_styling()
                    acc = [styler(element.text)]
                    acc.extend([child._md_format(styler) for child in element.children])
                    acc.append(_as_none(element.tail))
                    result.append(_strip_left_spaces("".join(x for x in acc if x)))
                sval = " ".join(result)
                return f"{filler}{sval.strip()}{filler}".replace("\\\n", " ").replace(
                    "\n", " "
                )
            return f"{filler}{str(val).strip()}{filler}".replace("\\\n", " ").replace(
                "\n", " "
            )

        def fill_cell(pos: int, cell_val: Any, filler: str = " ") -> str:
            sval = format_cell(cell_val, filler)
            step = sizer[pos] - len(sval)
            if step > 0:
                return sval + filler * step
            return sval

        def fill_line(cell_values: list[Any], filler: str = " ") -> list[str]:
            return [
                fill_cell(pos, cell_val, filler)
                for pos, cell_val in enumerate(cell_values)
            ]

        self.optimize_width()
        if not self.height:
            return ""
        sizer = {i: 3 for i in range(self.width)}  # noqa: C420
        safe_global = _copy_global()
        for idx in range(self.height):
            for i, val in enumerate(self.get_row_sub_elements(idx)):
                size = len(format_cell(val))
                if size > sizer[i]:
                    sizer[i] = size
        _restore_global(safe_global)
        result = []
        result.append(bars(fill_line(self.get_row_sub_elements(0))))
        result.append(bars(fill_line(["-"] * self.width, "-")))
        for idx in range(1, self.height):
            result.append(bars(fill_line(self.get_row_sub_elements(idx))))
        result.append("")
        return "\n".join(result)

    def _md_collect(self) -> list[str]:
        if content := self._md_format():
            return [content + "\n"]
        return []



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2012 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Cache classes for Table and Row."""

from __future__ import annotations

from bisect import bisect_left, insort
from typing import TYPE_CHECKING

from .element import xpath_compile

if TYPE_CHECKING:
    from lxml.etree import XPath

    from .cell import Cell
    from .column import Column
    from .row import Row
    from .table import Table

_XP_ROW_IDX = xpath_compile(
    "(table:table-row|table:table-rows/table:table-row|"
    "table:table-header-rows/table:table-row|"
    "table:table-row-group/child::table:table-row)[$idx]"
)
_XP_COLUMN_IDX = xpath_compile(
    "(table:table-column|table:table-columns/table:table-column|"
    "table:table-header-columns/table:table-column)[$idx]"
)
_XP_CELL_IDX = xpath_compile("(table:table-cell|table:covered-table-cell)[$idx]")


def _make_cache_map(idx_repeated_seq: list[tuple[int, int]]) -> list[int]:
    """Build the initial cache map of the table."""
    cache_map: list[int] = []
    for odf_idx, repeated in idx_repeated_seq:
        cache_map = _insert_map_once(cache_map, odf_idx, repeated)
    return cache_map


def _insert_map_once(
    cache_map: list[int],
    odf_idx: int,
    repeated: int,
) -> list[int]:
    """Add an item (cell or row) to the map.

    Args:
        cache_map: Cache map.
        odf_idx: Index in ODF XML.
        repeated: Repeated value of item, 1 or more.

    odf_idx is NOT position (col or row), neither raw XML position, but ODF index
    """
    repeated = repeated or 1
    if odf_idx > len(cache_map):
        raise IndexError
    if odf_idx > 0:
        before = cache_map[odf_idx - 1]
    else:
        before = -1
    juska = before + repeated  # aka max position value for item
    if odf_idx == len(cache_map):
        insort(cache_map, juska)
        return cache_map
    new_map = cache_map[:odf_idx]
    new_map.append(juska)
    new_map.extend([(x + repeated) for x in cache_map[odf_idx:]])
    return new_map


def _erase_map_once(cache_map: list[int], odf_idx: int) -> list[int]:
    """Remove an item (cell or row) from the map.

    Args:
        cache_map: Cache map.
        odf_idx: Index in ODF XML.
    """
    if odf_idx >= len(cache_map):
        raise IndexError
    if odf_idx > 0:
        before = cache_map[odf_idx - 1]
    else:
        before = -1
    current = cache_map[odf_idx]
    repeated = current - before
    cache_map = cache_map[:odf_idx] + [(x - repeated) for x in cache_map[odf_idx + 1 :]]
    return cache_map


def _insert_item_in_vault(
    position: int,
    item: Cell | Row | Column,
    idx: int,
    current_item: Cell | Row | Column,
    vault: Row | Table,
    vault_map: list[int],
) -> tuple[list[int], Cell | Row | Column]:
    repeated = item.repeated or 1
    target_idx = vault.index(current_item)
    current_cache = vault_map[idx]
    if idx > 0:
        before_cache = vault_map[idx - 1]
    else:
        before_cache = -1
    current_pos = before_cache + 1
    current_repeated = current_cache - before_cache
    repeated_before = position - current_pos
    repeated_after = current_repeated - repeated_before
    new_item = item.clone
    if repeated_before >= 1:
        current_item._set_repeated(repeated_before)
        vault.insert(new_item, position=target_idx + 1)
        after_item = current_item.clone
        after_item._set_repeated(repeated_after)
        vault.insert(after_item, position=target_idx + 2)
    else:
        # only insert new cell
        vault.insert(new_item, position=target_idx)
    # update cache
    if repeated_before >= 1:
        emap = _erase_map_once(vault_map, idx)
        emap = _insert_map_once(emap, idx, repeated_before)
        emap = _insert_map_once(emap, idx + 1, repeated)
        emap = _insert_map_once(emap, idx + 2, repeated_after)
    else:
        emap = _insert_map_once(vault_map, idx, repeated)
    return emap, new_item


def _set_item_in_vault(
    position: int,
    item: Cell | Row | Column,
    idx: int,
    current_item: Cell | Row | Column,
    vault: Row | Table,
    vault_map: list[int],
    vault_scheme: XPath,
    clone: bool = True,
) -> tuple[list[int], Cell | Row | Column]:
    """Set the item (cell, row) in its vault (row, table), updating the cache
    map.

    Args:
        position: Position of the item.
        item: The item to set.
        idx: Index in ODF XML.
        current_item: Current item.
        vault: Vault.
        vault_map: Vault map.
        vault_scheme: Vault scheme.
        clone: Whether to clone the item.
    """
    repeated = item.repeated or 1
    target_idx = vault.index(current_item)
    current_cache = vault_map[idx]
    if idx > 0:
        before_cache = vault_map[idx - 1]
    else:
        before_cache = -1
    current_pos = before_cache + 1
    current_repeated = current_cache - before_cache
    repeated_before = position - current_pos
    repeated_after = current_repeated - repeated_before - repeated
    if repeated_before >= 1:
        # Update repetition
        current_item._set_repeated(repeated_before)
        target_idx += 1
    else:
        # Replacing the first occurrence
        vault.delete(current_item)
    # Insert new element
    if clone:
        new_item = item.clone
    else:
        new_item = item
    vault.insert(new_item, position=target_idx)
    # Insert the remaining repetitions
    if repeated_after >= 1:
        after_item = current_item.clone
        after_item._set_repeated(repeated_after)
        vault.insert(after_item, position=target_idx + 1)
    # setting a repeated item !
    if repeated_after < 0:
        # deleting some overlapped items
        deleting = repeated_after
        while deleting < 0:
            delete_item = vault._get_element_idx2(vault_scheme, target_idx + 1)
            if delete_item is None:
                break
            is_repeated = delete_item.repeated or 1  # type: ignore
            is_repeated += deleting
            if is_repeated > 1:
                delete_item._set_repeated(is_repeated)  # type: ignore
            else:
                vault.delete(delete_item)
            deleting = is_repeated
    # update cache
    # remove existing
    emap = _erase_map_once(vault_map, idx)
    # add before if any:
    if repeated_before >= 1:
        emap = _insert_map_once(emap, idx, repeated_before)
        idx += 1
    # add our slot
    emap = _insert_map_once(emap, idx, repeated)
    # add after if any::
    if repeated_after >= 1:
        idx += 1
        emap = _insert_map_once(emap, idx, repeated_after)
    if repeated_after < 0:
        idx += 1
        while repeated_after < 0:
            if idx < len(emap):
                emap = _erase_map_once(emap, idx)
            repeated_after += 1
    return emap, new_item


def _delete_item_in_vault(
    idx: int,
    current_item: Cell | Row | Column,
    vault: Row | Table,
    vault_map: list[int],
) -> list[int]:
    current_cache = vault_map[idx]
    if idx > 0:
        before_cache = vault_map[idx - 1]
    else:
        before_cache = -1
    # current_pos = before_cache + 1
    current_repeated = current_cache - before_cache
    new_repeated = current_repeated - 1
    if new_repeated >= 1:
        current_item._set_repeated(new_repeated)
        emap = vault_map[:idx] + [(x - 1) for x in vault_map[idx:]]
    else:
        # actual erase
        vault.delete(current_item)
        emap = vault_map[:idx] + [(x - 1) for x in vault_map[idx + 1 :]]
    return emap


class RowCache:
    """Cache for Row (internal)."""

    __slots__ = ("cell_elements", "cell_map")

    def __init__(self) -> None:
        self.cell_map: list[int] = []
        self.cell_elements: dict[int, Cell] = {}

    def __str__(self) -> str:
        return f"RC cell:{self.cell_map!r}"

    @classmethod
    def copy(cls, source: RowCache) -> RowCache:
        rc = cls()
        rc.cell_map = source.cell_map[:]
        return rc

    def width(self) -> int:
        """Get the number of expected cells in the row, i.e. addition
        repetitions.

        Returns:
            int: The number of expected cells in the row.
        """
        try:
            return self.cell_map[-1] + 1
        except IndexError:
            return 0

    def clear_cell_indexes(self) -> None:
        self.cell_elements = {}

    def cell_idx(self, position: int) -> int | None:
        """Find cell index in the map from the position."""
        idx = bisect_left(self.cell_map, position)
        if idx < len(self.cell_map):
            return idx
        return None

    def cell_map_length(self) -> int:
        return len(self.cell_map)

    def cached_cell(self, idx: int) -> Cell | None:
        """Retrieve Cell in cache."""
        return self.cell_elements.get(idx)

    def store_cell(self, cell: Cell, idx: int) -> None:
        """Store Cell in cache."""
        self.cell_elements[idx] = cell

    def insert_cell_map_once(self, repeated: int) -> None:
        self.cell_map = _insert_map_once(self.cell_map, len(self.cell_map), repeated)

    # def erase_cell_map_once(self, odf_idx: int) -> None:
    #     self.cell_map = _erase_map_once(self.cell_map, odf_idx)

    def make_cell_map(self, idx_repeated_sequence: list[tuple[int, int]]) -> None:
        self.cell_map = _make_cache_map(idx_repeated_sequence)

    def set_cell_in_cache(
        self,
        x: int,
        cell: Cell,
        vault: Row,
        clone: bool,
    ) -> Cell:
        idx = self.cell_idx(x)
        if idx is None:
            raise ValueError
        current_cached_cell: Cell | None = self.cached_cell(idx)
        if current_cached_cell is None:
            current_cached_cell = vault._get_element_idx2(_XP_CELL_IDX, idx)  # type: ignore[assignment]
        if not current_cached_cell:
            raise ValueError  # pragma: nocover
        self.clear_cell_indexes()
        emap, new_cell = _set_item_in_vault(
            x,
            cell,
            idx,
            current_cached_cell,
            vault,
            self.cell_map,
            _XP_CELL_IDX,
            clone,
        )
        self.cell_map = emap
        return new_cell  # type: ignore[return-value]

    def insert_cell_in_cache(
        self,
        x: int,
        cell: Cell,
        vault: Row,
    ) -> Cell:
        idx = self.cell_idx(x)
        if idx is None:
            raise ValueError
        current_cached_cell: Cell | None = self.cached_cell(idx)
        if current_cached_cell is None:
            current_cached_cell = vault._get_element_idx2(_XP_CELL_IDX, idx)  # type: ignore[assignment]
        if not current_cached_cell:
            raise ValueError  # pragma: nocover
        self.clear_cell_indexes()
        emap, new_cell = _insert_item_in_vault(
            x,
            cell,
            idx,
            current_cached_cell,
            vault,
            self.cell_map,
        )
        self.cell_map = emap
        return new_cell  # type: ignore[return-value]

    def delete_cell_in_cache(
        self,
        x: int,
        vault: Row,
    ) -> None:
        idx = self.cell_idx(x)
        if idx is None:
            raise ValueError
        current_cached_cell: Cell | None = self.cached_cell(idx)
        if current_cached_cell is None:
            current_cached_cell = vault._get_element_idx2(_XP_CELL_IDX, idx)  # type: ignore[assignment]
        if not current_cached_cell:
            raise ValueError  # pragma: nocover
        self.clear_cell_indexes()
        emap = _delete_item_in_vault(
            idx,
            current_cached_cell,
            vault,
            self.cell_map,
        )
        self.cell_map = emap


class TableCache:
    """Cache for Table (internal)."""

    __slots__ = ("col_elements", "col_map", "row_elements", "row_map")

    def __init__(self) -> None:
        self.row_map: list[int] = []
        self.col_map: list[int] = []
        self.row_elements: dict[int, Row] = {}
        self.col_elements: dict[int, Column] = {}

    def __str__(self) -> str:
        return f"TC row:{self.row_map!r} col:{self.col_map!r}"

    @classmethod
    def copy(cls, source: TableCache) -> TableCache:
        tc = cls()
        tc.row_map = source.row_map[:]
        tc.col_map = source.col_map[:]
        return tc

    def height(self) -> int:
        """Get the current height of the table.

        Returns:
            int: The current height of the table.
        """
        try:
            return self.row_map[-1] + 1
        except IndexError:
            return 0

    def width(self) -> int:
        """Get the current width of the table, measured on columns.

        Rows may have different widths, use the Table API to ensure width
        consistency.

        Returns:
            int: The current width of the table.
        """
        try:
            return self.col_map[-1] + 1
        except IndexError:
            return 0

    def clear_row_indexes(self) -> None:
        self.row_elements = {}

    def clear_col_indexes(self) -> None:
        self.col_elements = {}

    def row_idx(self, position: int) -> int | None:
        """Find row index in the map from the position."""
        idx = bisect_left(self.row_map, position)
        if idx < len(self.row_map):
            return idx
        return None

    def col_idx(self, position: int) -> int | None:
        """Find column index in the map from the position."""
        idx = bisect_left(self.col_map, position)
        if idx < len(self.col_map):
            return idx
        return None

    def col_map_length(self) -> int:
        return len(self.col_map)

    def cached_row(self, idx: int) -> Row | None:
        """Retrieve Row in cache."""
        return self.row_elements.get(idx)

    def cached_col(self, idx: int) -> Column | None:
        """Retrieve Column in cache."""
        return self.col_elements.get(idx)

    def store_row(self, row: Row, idx: int) -> None:
        """Store Row in cache."""
        self.row_elements[idx] = row

    def store_col(self, col: Column, idx: int) -> None:
        """Store Column in cache."""
        self.col_elements[idx] = col

    def insert_row_map_once(self, repeated: int) -> None:
        self.row_map = _insert_map_once(self.row_map, len(self.row_map), repeated)

    # def erase_row_map_once(self, odf_idx: int) -> None:
    #     self.row_map = _erase_map_once(self.row_map, odf_idx)

    def insert_col_map_once(self, repeated: int) -> None:
        self.col_map = _insert_map_once(self.col_map, len(self.col_map), repeated)

    # def erase_col_map_once(self, odf_idx: int) -> None:
    #     self.col_map = _erase_map_once(self.col_map, odf_idx)

    def make_row_map(self, idx_repeated_sequence: list[tuple[int, int]]) -> None:
        self.row_map = _make_cache_map(idx_repeated_sequence)

    def make_col_map(self, idx_repeated_sequence: list[tuple[int, int]]) -> None:
        self.col_map = _make_cache_map(idx_repeated_sequence)

    def set_row_in_cache(
        self,
        y: int,
        row: Row,
        vault: Table,
        clone: bool,
    ) -> Row:
        idx = self.row_idx(y)
        if idx is None:
            raise ValueError
        current_cached_row: Row | None = self.cached_row(idx)
        if current_cached_row is None:
            current_cached_row = vault._get_element_idx2(_XP_ROW_IDX, idx)  # type: ignore[assignment]
        if not current_cached_row:
            raise ValueError  # pragma: nocover
        self.clear_row_indexes()
        emap, new_row = _set_item_in_vault(
            y,
            row,
            idx,
            current_cached_row,
            vault,
            self.row_map,
            _XP_ROW_IDX,
            clone,
        )
        self.row_map = emap
        return new_row  # type: ignore[return-value]

    def insert_row_in_cache(
        self,
        y: int,
        row: Row,
        vault: Table,
    ) -> Row:
        idx = self.row_idx(y)
        if idx is None:
            raise ValueError
        current_cached_row: Row | None = self.cached_row(idx)
        if current_cached_row is None:
            current_cached_row = vault._get_element_idx2(_XP_ROW_IDX, idx)  # type: ignore[assignment]
        if not current_cached_row:
            raise ValueError  # pragma: nocover
        self.clear_row_indexes()
        emap, new_row = _insert_item_in_vault(
            y,
            row,
            idx,
            current_cached_row,
            vault,
            self.row_map,
        )
        self.row_map = emap
        return new_row  # type: ignore[return-value]

    def delete_row_in_cache(
        self,
        y: int,
        vault: Table,
    ) -> None:
        idx = self.row_idx(y)
        if idx is None:
            raise ValueError
        current_cached_row: Row | None = self.cached_row(idx)
        if current_cached_row is None:
            current_cached_row = vault._get_element_idx2(_XP_ROW_IDX, idx)  # type: ignore[assignment]
        if not current_cached_row:
            raise ValueError  # pragma: nocover
        self.clear_row_indexes()
        emap = _delete_item_in_vault(
            idx,
            current_cached_row,
            vault,
            self.row_map,
        )
        self.row_map = emap

    def set_col_in_cache(
        self,
        x: int,
        column: Column,
        vault: Table,
    ) -> Column:
        idx = self.col_idx(x)
        if idx is None:
            raise ValueError
        current_cached_col: Column | None = self.cached_col(idx)
        if current_cached_col is None:
            current_cached_col = vault._get_element_idx2(_XP_COLUMN_IDX, idx)  # type: ignore[assignment]
        if not current_cached_col:
            raise ValueError  # pragma: nocover
        self.clear_col_indexes()
        emap, new_col = _set_item_in_vault(
            x,
            column,
            idx,
            current_cached_col,
            vault,
            self.col_map,
            _XP_COLUMN_IDX,
        )
        self.col_map = emap
        return new_col  # type: ignore[return-value]

    def insert_col_in_cache(
        self,
        x: int,
        column: Column,
        vault: Table,
    ) -> Column:
        idx = self.col_idx(x)
        if idx is None:
            raise ValueError
        current_cached_col: Column | None = self.cached_col(idx)
        if current_cached_col is None:
            current_cached_col = vault._get_element_idx2(_XP_COLUMN_IDX, idx)  # type: ignore[assignment]
        if not current_cached_col:
            raise ValueError  # pragma: nocover
        self.clear_col_indexes()
        emap, new_col = _insert_item_in_vault(
            x,
            column,
            idx,
            current_cached_col,
            vault,
            self.col_map,
        )
        self.col_map = emap
        return new_col  # type: ignore[return-value]

    def delete_col_in_cache(
        self,
        x: int,
        vault: Table,
    ) -> None:
        idx = self.col_idx(x)
        if idx is None:
            raise ValueError
        current_cached_col: Column | None = self.cached_col(idx)
        if current_cached_col is None:
            current_cached_col = vault._get_element_idx2(_XP_COLUMN_IDX, idx)  # type: ignore[assignment]
        if not current_cached_col:
            raise ValueError  # pragma: nocover
        self.clear_col_indexes()
        emap = _delete_item_in_vault(
            idx,
            current_cached_col,
            vault,
            self.col_map,
        )
        self.col_map = emap



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2012 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: David Versmisse <david.versmisse@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          Romain Gauthier <romain@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
"""Coordinate translation utilities for ODF documents.

This module provides functions for converting between different coordinate
representations used in ODF tables, such as "A1" style, (row, column) tuples,
and numeric indices.
"""

from __future__ import annotations

from .isiterable import isiterable


def translate_from_any(x: str | int, length: int, idx: int) -> int:
    """Translate a coordinate from any format to a 0-based integer index.

    Args:
        x: The coordinate, which can be a 1-based integer or a string.
        length: The length of the axis for handling negative indices.
        idx: The index in the coordinate tuple to use (0 for x, 1 for y).

    Returns:
        int: The 0-based integer coordinate.

    Raises:
        TypeError: If the input value is not a string or integer.
    """
    if isinstance(x, str):
        value_int = convert_coordinates(x)[idx]
        if value_int is None:
            raise TypeError(f"Wrong value: {x!r}")
    elif isinstance(x, int):
        value_int = x
    else:
        raise TypeError(f"Wrong value: {x!r}")
    if value_int < 0:
        return increment(value_int, length)
    return value_int


def alpha_to_digit(alpha: str) -> int:
    """Translates a column name from alphabetic to a 0-based numeric index.

    For example, "A" becomes 0, "B" becomes 1, and "AB" becomes 27.

    Args:
        alpha: The alphabetic column name.

    Returns:
        int: The 0-based numeric index of the column.

    Raises:
        ValueError: If the input string contains non-alphabetic characters.
    """
    if isinstance(alpha, int):
        return alpha
    if not alpha.isalpha():
        raise ValueError(f"Column value {alpha!r} is malformed")
    column = 0
    for c in alpha.lower():
        val = ord(c) - ord("a") + 1
        column = column * 26 + val
    return column - 1


def digit_to_alpha(digit: int | str) -> str:
    """Translates a 0-based column index to its alphabetic representation.

    For example, 0 becomes "A", 1 becomes "B", and 27 becomes "AB".

    Args:
        digit: The 0-based numeric index of the column.

    Returns:
        str: The alphabetic representation of the column.

    Raises:
        TypeError: If the input is not an integer.
    """
    if isinstance(digit, str) and digit.isalpha():
        return digit
    if not isinstance(digit, int):
        raise TypeError(f'column number "{digit}" is invalid')
    digit += 1
    column = ""
    while digit:
        column = chr(65 + ((digit - 1) % 26)) + column
        digit = (digit - 1) // 26
    return column


def increment(value: int, step: int) -> int:
    """Adjusts a negative index to a positive one based on a step.

    This is used to handle negative indexing in table coordinates.

    Args:
        value: The index value, which may be negative.
        step: The total size of the dimension (e.g., table width or height).

    Returns:
        int: The adjusted, non-negative index.
    """
    while value < 0:
        if step == 0:
            return 0
        value += step
    return value


def convert_coordinates(
    obj: tuple | list | str,
) -> tuple[int | None, ...]:
    """Translates various coordinate formats into a tuple of 0-based integers.

    This function can handle formats like "A1", "A1:C3", or tuples like (0, 0).
    A single cell coordinate is returned as a (column, row) tuple. An area is
    returned as a (col1, row1, col2, row2) tuple.

    Args:
        obj: The coordinate representation to convert.

    Returns:
        tuple[int | None, ...]: A tuple of 0-based integer coordinates.

    Raises:
        TypeError: If the input object is of an unsupported type.
        ValueError: If the coordinate string is malformed.

    Examples:
        >>> convert_coordinates("D3")
        (3, 2)
        >>> convert_coordinates((1, 2))
        (1, 2)
        >>> convert_coordinates("A1:B3")
        (0, 0, 1, 2)
    """
    # By (1, 2) ?
    if isiterable(obj):
        return tuple(obj)
    # Or by 'B3' notation ?
    if not isinstance(obj, str):
        raise TypeError(f'Bad coordinates type: "{type(obj)}"')
    coordinates = []
    for coord in [x.strip() for x in obj.split(":", 1)]:
        # First "A"
        alpha = ""
        for c in coord:
            if c.isalpha():
                alpha += c
            else:
                break
        try:
            column = alpha_to_digit(alpha)
        except ValueError:
            # raise ValueError, 'coordinates "%s" malformed' % obj
            # maybe '1:4' table row coordinates
            column = None
        coordinates.append(column)
        # Then "1"
        try:
            line = int(coord[len(alpha) :]) - 1
        except ValueError:
            # raise ValueError, 'coordinates "%s" malformed' % obj
            # maybe 'A:C' row coordinates
            line = None
        if line and line < 0:
            raise ValueError(f"Coordinates {obj!r} malformed")
        coordinates.append(line)
    return tuple(coordinates)



#!/usr/bin/env python
"""Demo of quick introspecting of a document's elements.

The body object of a document is a mapping of an XML tree from which we
can access other elements we are looking for (parent, children)."""

import os
import sys
from pathlib import Path
from pprint import pformat
from typing import Any

from odfdo import Document

_DOC_SEQUENCE = 480
DATA = Path(__file__).parent / "data"
# ODF export of Wikipedia article Hitchhiker's Guide to the Galaxy (CC-By-SA) :
SOURCE = "collection2.odt"


def read_source_document() -> Document:
    """Return the source Document."""
    try:
        source = sys.argv[1]
    except IndexError:
        source = DATA / SOURCE
    return Document(source)


def analyser(document: Document) -> dict[str, Any]:
    """Return information from an element of the document."""

    result: dict[str, Any] = {}
    # Elements are part of an XML tree:
    paragraph = document.body.get_paragraph(position=42)

    result["tag"] = paragraph.tag
    result["attributes"] = paragraph.attributes
    result["str"] = str(paragraph)
    result["parent"] = paragraph.parent
    result["children"] = paragraph.children
    result["serialize"] = paragraph.serialize(pretty=True)

    print("Informations about the paragraph:")
    print(pformat(result))
    return result


def main() -> None:
    document = read_source_document()
    result = analyser(document)
    test_unit(result)


def test_unit(result: dict[str, Any]) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    assert result["tag"] == "text:p"
    assert repr(result["parent"]) == "<NoteBody tag=text:note-body>"
    assert repr(result["children"]) == "[<Span tag=text:span>]"


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Authors (odfdo project): jerome.dumonteil@gmail.com
from __future__ import annotations

import io
import subprocess
import sys
from decimal import Decimal
from pathlib import Path

import pytest

from odfdo.document import Document
from odfdo.scripts import from_csv
from odfdo.scripts.from_csv import main as main_script
from odfdo.scripts.from_csv import main_from_csv, parse_cli_args

SCRIPT = Path(from_csv.__file__)


def run_params(params: list):
    command = [sys.executable, SCRIPT] + params
    proc = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = proc.communicate()
    return out, err, proc.returncode


def test_from_csv_no_param():
    params = []
    _out, _err, exitcode = run_params(params)
    assert exitcode == 1


# direct access to internal function


def test_from_csv_2_no_param_on_main_function(monkeypatch):
    with pytest.raises(Exception) as result:
        monkeypatch.setattr(sys, "argv", [])
        main_script()
        assert result.value.code >= 1


def test_from_csv_2_no_param(monkeypatch):
    with pytest.raises(Exception) as result:
        params = parse_cli_args([])
        main_from_csv(params)
        assert result.value.code >= 1


def test_from_csv_2_version(capsys):
    with pytest.raises(SystemExit) as result:
        parse_cli_args(["--version"])
        assert result.value.code == 0
    captured = capsys.readouterr()

    assert "odfdo-from-csv v3" in captured.out


def test_from_csv_2_no_file():
    params = parse_cli_args(["-i", "none_file"])

    with pytest.raises(FileNotFoundError) as result:
        main_from_csv(params)
        assert result.value.code == 1


def test_from_csv_2_text1(capsysbinary, samples):
    source = samples("text1.csv")
    params = parse_cli_args(["-i", str(source)])

    main_from_csv(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    table = document.body.tables[0]
    assert table.name == "table"
    assert table.get_row_values(0) == ["foo1", "foo2"]
    assert table.get_row_values(1) == [1, 2]


def test_from_csv_2_text2(capsysbinary, samples):
    source = samples("text2.csv")
    params = parse_cli_args(["-i", str(source), "-t", "sheet"])

    main_from_csv(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    table = document.body.tables[0]
    assert table.name == "sheet"
    # buggy windows envirnment not beeing utf-8:
    deci_bug = "dÃ©cimal"
    # to fix: 'dÃ©cimal'.encode('latin1').decode()
    case1 = ["foo1", "foo2", "txt content", "décimal"]
    case2 = ["foo1", "foo2", "txt content", deci_bug]
    assert table.get_row_values(0) in (case1, case2)
    assert table.get_row_values(1) == [1, 2, "some text with, comma", Decimal("-3.14")]
    assert table.get_row_values(2) == [3, 4, "text with space", Decimal("0.01")]



#!/usr/bin/env python
"""Example of methods and properties to analyse a document.

These methods or properties return a list of elements:

    - `body.headers`
    - `body.images`
    - `body.paragraphs`
    - `body.get_links()`
    - `body.get_notes()`
    - `body.tables` (alias: body.sheets)
    - `body.get_paragraphs(content)`
"""

# Expected result on stdout:
# number of headings: 29
# number of images stored: 0
# number of paragraphs: 175
# number of links (URLs): 352
# number of footnotes: 49
# number of tables: 0
# Paragraphs with 'Fish': 4
# Paragraphs with 'answer' and '42': 1

import os
import sys
from pathlib import Path

from odfdo import Document

_DOC_SEQUENCE = 75
DATA = Path(__file__).parent / "data"
SOURCE = "collection2.odt"


def read_source_document() -> Document:
    """Return the source Document."""
    try:
        source = sys.argv[1]
    except IndexError:
        source = DATA / SOURCE
    return Document(source)


def analysis(document: Document) -> dict[str, int]:
    """Returns some statistics about the document."""
    result: dict[str, int] = {
        "headings": 0,
        "images": 0,
        "paragraphs": 0,
        "links": 0,
        "footnotes": 0,
        "tables": 0,
        "fish": 0,
        "answer": 0,
    }

    # The body object is an XML element from which we can access one or several
    # other elements we are looking for.
    body = document.body

    # Should you need to access all elements of a kind, there are the
    # get_xxxs methods, where xxx can be paragraph, heading, list, table, ...
    # Methods without parameters are accessible through properties.

    # Some examples, that you can check against actual content of the odt file:
    # See how complex is our wikipedia documents:
    result["headings"] = len(body.headers)
    result["images"] = len(body.images)
    result["paragraphs"] = len(body.paragraphs)
    result["links"] = len(body.get_links())
    result["footnotes"] = len(body.get_notes())
    # Our sample document has no table:
    # print("number of tables:", len(body.get_tables()))
    result["tables"] = len(body.tables)
    # sheets is an alias of tables:
    # result["tables"] = len(body.sheets)

    # Each get_xxx_list method provides parameters for filtering the results.
    # For example headings can be listed by level, annotations by creator, etc.
    # Almost all of them accept filtering by style and content using a regular
    # expressions.
    result["fish"] = len(body.get_paragraphs(content=r"Fish"))
    result["answer"] = len(body.get_paragraphs(content=r"answer.*42"))

    return result


def display_analysis(stats: dict[str, int]) -> None:
    """Print the stats on stdout."""
    print(f"number of headings: {stats['headings']}")
    print(f"number of images stored: {stats['images']}")
    print(f"number of paragraphs: {stats['paragraphs']}")
    print(f"number of links (URLs): {stats['links']}")
    print(f"number of footnotes: {stats['footnotes']}")
    print(f"number of tables: {stats['tables']}")
    print(f"Paragraphs with 'Fish': {stats['fish']}")
    print(f"Paragraphs with 'answer' and '42': {stats['answer']}")


def main() -> None:
    document = read_source_document()
    stats = analysis(document)
    display_analysis(stats)
    test_unit(stats)


def test_unit(stats: dict[str, int]) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    assert stats["headings"] == 29
    assert stats["images"] == 0
    assert stats["paragraphs"] == 175
    assert stats["links"] == 352
    assert stats["footnotes"] == 49
    assert stats["tables"] == 0
    assert stats["fish"] == 4
    assert stats["answer"] == 1


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Classes related to "style:page-layout"."""

from __future__ import annotations

from typing import Any, ClassVar, Union, cast

from .element import Element, PropDef, PropDefBool, register_element_class
from .style_base import PropDict, StyleBase
from .style_props import StyleProps
from .style_utils import (
    _set_background,
)


class StylePageLayout(StyleProps):
    """The "style:page-layout" element represents the styles that specify the
    formatting properties of a page.

    The "style:page-layout" element is usable within the following element:
    "office:automatic-styles".
    """

    # The "style:page-layout" element has the following attributes:
    # style:name
    # style:page-usage

    _tag: str = "style:page-layout"
    _properties: tuple[PropDef | PropDefBool, ...] = (PropDef("name", "style:name"),)
    PAGE_USAGE: ClassVar = {"all", "left", "right", "mirrored"}

    def __init__(
        self,
        name: str | None = None,
        page_usage: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a StylePageLayout element.

        The `name` is not mandatory at creation but becomes required when
        inserting into a document as an automatic style.

        Args:
            name: The name of the page layout style.
            page_usage: The type of pages the layout applies to.
                Allowed values are "all" (default), "left", "mirrored", "right".
            **kwargs: Additional keyword arguments for the parent `Element` class.
        """
        self._family = "page-layout"
        tag_or_elem = kwargs.get("tag_or_elem")
        if tag_or_elem is None:
            kwargs["tag"] = "style:page-layout"
        Element.__init__(self, **kwargs)
        if self._do_init:
            kwargs.pop("tag", None)
            kwargs.pop("tag_or_elem", None)
            if name:
                self.name = name
            if page_usage:
                self.page_usage = page_usage

    @property
    def family(self) -> str | None:
        """Get the family of the style.

        For `StylePageLayout`, this is always "page-layout".

        Returns:
            str | None: The family name.
        """
        return self._family

    @family.setter
    def family(self, family: str | None) -> None:
        """Setter for the family property (no-op as family is fixed)."""
        pass

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} family={self.family} name={self.name}>"

    @property
    def page_usage(self) -> str:
        """Get the `style:page-usage` attribute.

        Specifies the type of pages that a page master should generate.

        Returns:
            str: The page usage type. Defaults to "all".
        """
        return self._get_attribute_str_default("style:page-usage", "all")

    @page_usage.setter
    def page_usage(self, page_usage: str | None) -> None:
        """Set the `style:page-usage` attribute.

        Args:
            page_usage: The page usage type. Allowed values are
                "all", "left", "right", "mirrored". Invalid values default to "all".
        """
        if page_usage not in self.PAGE_USAGE:
            page_usage = "all"
        self._set_attribute_str_default("style:page-usage", page_usage, "all")

    def get_properties(self, area: str | None = "page-layout") -> PropDict | None:
        """Retrieve the page-layout properties of the `StylePageLayout`.

        Args:
            area: The area for which to retrieve properties.
                (Parameter is kept for compatibility but internally fixed to "page-layout").

        Returns:
            dict[str, str | dict] | None: A dictionary mapping property names to their values.
        """
        return super().get_properties(area="page-layout")

    def set_properties(
        self,
        properties: PropDict | None = None,
        style: StyleBase | None = None,
        area: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Set the properties for the page-layout.

        Properties can be provided either as a dictionary, as a `StyleBase` object,
        or as keyword arguments. The `area` is internally fixed to "page-layout".

        Args:
            properties: A dictionary of properties to set.
            style: A `StyleBase` object from which to copy properties.
            area: The area for which to set properties.
                (Parameter is kept for compatibility but internally fixed to "page-layout").
            **kwargs: Additional keyword arguments for properties to set.
        """
        return super().set_properties(
            properties=properties, style=style, area="page-layout", **kwargs
        )

    def del_properties(
        self,
        properties: list[str] | None = None,
        area: str | None = None,
    ) -> None:
        """Delete specific properties from the page-layout.

        Args:
            properties: A list of property names to delete.
            area: The area from which to delete properties.
                (Parameter is kept for compatibility but internally fixed to "page-layout").
        """
        return super().del_properties(properties=properties, area="page-layout")

    def set_background(
        self,
        color: str | None = None,
        url: str | None = None,
        position: str | None = "center",
        repeat: str | None = None,
        opacity: str | int | None = None,
        filter: str | None = None,  # noqa: A002
    ) -> None:
        """Set the background properties for the page layout.

        This can configure a background color or an image. If no arguments are
        provided, any existing background is removed.

        Args:
            color: The background color in '#RRGGBB' format.
            url: The URL of a background image.
            position: The position of the background image. Can be
                "left", "center", "right", "top", "bottom", or a combination
                of two (e.g., "top center"). Defaults to "center".
            repeat: How the background image repeats. Can be
                "no-repeat", "repeat", or "stretch".
            opacity: The opacity of the background image as
                a percentage integer (0-100).
            filter: An application-specific filter name for the background image.
        """
        _set_background(self, color, url, position, repeat, opacity, filter)

    def get_header_style(self) -> StyleBase | None:
        """Get the `style:header-style` element within the page layout.

        Returns:
            StyleBase | None: The `StyleBase` instance representing the header
                style, or `None` if no header style is defined.
        """
        return cast(Union[None, StyleBase], self.get_element("style:header-style"))

    def set_header_style(self, new_style: StyleBase) -> None:
        """Set or replace the `style:header-style` element within the page layout.

        Args:
            new_style: The new header style to set.
        """
        header_style = self.get_header_style()
        if header_style is not None:
            self.delete(header_style)
        self.append(new_style)

    def get_footer_style(self) -> StyleBase | None:
        """Get the `style:footer-style` element within the page layout.

        Returns:
            StyleBase | None: The `StyleBase` instance representing the footer
                style, or `None` if no footer style is defined.
        """
        return cast(Union[None, StyleBase], self.get_element("style:footer-style"))

    def set_footer_style(self, new_style: StyleBase) -> None:
        """Set or replace the `style:footer-style` element within the page layout.

        Args:
            new_style: The new footer style to set.
        """
        footer_style = self.get_footer_style()
        if footer_style is not None:
            self.delete(footer_style)
        self.append(new_style)


StylePageLayout._define_attribut_property()


register_element_class(StylePageLayout)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Language tag validation utility based on RFC 3066.

This module provides a function to validate language tags according to the
format defined in RFC 3066, which is commonly used for language identification
in ODF documents.
"""

from __future__ import annotations

from string import ascii_letters, digits


def is_RFC3066(lang: str) -> bool:
    """Checks if a string conforms to the RFC 3066 language tag format.

    Valid formats are "language" or "language-country", where "language" is a
    2 or 3-letter ASCII string, and "country" (and other subtags) are
    alphanumeric.

    Args:
        lang: The language tag string to validate.

    Returns:
        bool: True if the tag is valid, False otherwise.
    """

    def test_part1(part1: str) -> bool:
        if not 2 <= len(part1) <= 3:
            return False
        return all(x in ascii_letters for x in part1)

    def test_part2(part2: str) -> bool:
        return all(x in ascii_letters or x in digits for x in part2)

    if not lang or not isinstance(lang, str):
        return False
    if "-" not in lang:
        return test_part1(lang)
    parts = lang.split("-")
    if len(parts) > 3:
        return False
    if not test_part1(parts[0]):
        return False
    return all(test_part2(p) for p in parts[1:])



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>

from collections.abc import Iterable

import pytest

from odfdo.column import Column


@pytest.fixture
def column() -> Iterable[Column]:
    yield Column(default_cell_style="ce5", repeated=7, style="co1")


def test_get_column_default_cell_style(column):
    assert column.get_default_cell_style() == "ce5"


def test_get_column_default_cell_style_property(column):
    assert column.default_cell_style == "ce5"


def test_set_column_default_cell_style(column):
    column.set_default_cell_style("ce2")
    assert column.get_default_cell_style() == "ce2"
    column.set_default_cell_style(None)
    assert column.get_default_cell_style() is None


def test_set_column_default_cell_style_property(column):
    column.default_cell_style = "ce2"
    assert column.default_cell_style == "ce2"
    column.default_cell_style = None
    assert column.default_cell_style is None


def test_get_column_repeated(column):
    assert column.repeated == 7


def test_set_column_repeated(column):
    column.repeated = 99
    assert column.repeated == 99
    column.repeated = 1
    assert column.repeated is None
    column.repeated = 2
    assert column.repeated == 2
    column.repeated = None
    assert column.repeated is None


def test_get_column_style(column):
    assert column.style == "co1"


def test_set_column_style(column):
    column.style = "co2"
    assert column.style == "co2"
    column.style = None
    assert column.style is None



#!/usr/bin/env python
"""Create a short text document containing a list of items and a few sublists.
The code demonstrates several manipulations of the list and its items, then
displays the result to standard output."""

import os
from pathlib import Path

from odfdo import Document, List, ListItem

_DOC_SEQUENCE = 25
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "basic_sublist"
TARGET = "document.odt"


def save_new(document: Document, name: str) -> None:
    """Save a recipe result Document."""
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    new_path = OUTPUT_DIR / name
    print("Saving:", new_path)
    document.save(new_path, pretty=True)


def generate_document() -> Document:
    """Generate a basic document containing some lists and sublists."""
    document = Document("text")
    body = document.body
    body.clear()

    # Make a list:
    name_list = List(["Arthur", "Ford", "Trillian"])
    # It is possible to add an item to the list afterwards:
    item = ListItem("Marvin")
    name_list.append_item(item)

    # add list to the document:
    body.append(name_list)

    # Adding Sublist:
    # A sublist is simply a list as an item of another list:
    item.append(List(["Paranoid Android", "older than the universe"]))

    # See the result:
    print(document.get_formatted_text())
    # - Arthur
    # - Ford
    # - Trillian
    # - Marvin
    #
    #   - Paranoid Android
    #   - older than the universe

    # Inserting List Item
    # In case your forgot to insert an item:
    name_list.insert_item("some dolphins", position=1)

    # Or you can insert it before another item:
    marvin = name_list.get_item(content="Marvin")
    name_list.insert_item("Zaphod", before=marvin)
    # Or after:
    name_list.insert_item("and many others", after=marvin)

    # See the result:
    print(document.get_formatted_text())
    # - Arthur
    # - some dolphins
    # - Ford
    # - Trillian
    # - Zaphod
    # - Marvin
    #
    #   - Paranoid Android
    #   - older than the universe
    # - and many others
    #

    return document


def main() -> None:
    document = generate_document()
    test_unit(document)
    save_new(document, TARGET)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    assert document.get_formatted_text().strip() == (
        "- Arthur\n"
        "- some dolphins\n"
        "- Ford\n"
        "- Trillian\n"
        "- Zaphod\n"
        "- Marvin\n"
        "  \n"
        "  - Paranoid Android\n"
        "  - older than the universe\n"
        "- and many others"
    )


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Representation of the "settings.xml" part of the ODF document.

This part stores document-wide settings.
"""

from __future__ import annotations

from typing import Any, Union, cast

from .body import OfficeSettings
from .config_elements import ConfigItemSet
from .element import Element
from .xmlpart import XmlPart


class Settings(XmlPart):
    """Representation of the "settings.xml" part of the ODF document.

    This part stores document-wide settings.
    """

    def _get_body(self) -> OfficeSettings:
        body = cast(Union[None, OfficeSettings], self.get_element("//office:settings"))
        if isinstance(body, OfficeSettings):
            return body
        raise TypeError("No OfficeSettings found")  # pragma: nocover

    @property
    def odf_office_version(self) -> str:
        """Get the "office:version" value of the settings part.

        This value indicates the ODF version of the document settings.

        Returns:
            str: The "office:version" value, or an empty string if not found.
        """
        odsettings = cast(
            Union[None, Element], self.get_element("//office:document-settings")
        )
        # "office:version" should be always present
        if odsettings:
            return odsettings.get_attribute_string("office:version") or ""
        return ""  # pragma: nocover

    @property
    def config_item_sets(self) -> list[ConfigItemSet]:
        """Get a list of first-level ConfigItemSet elements within the
        settings.

        These represent top-level configuration item sets in the document's
        settings.

        Returns:
        list[ConfigItemSet]: A list of `ConfigItemSet` objects.
        """
        return cast(
            list[ConfigItemSet], self.body.get_elements("config:config-item-set")
        )

    def as_dict(self) -> dict[str, str | int | bool | dict[str, Any] | list[Any]]:
        """Serialize the settings content into a dictionary.

        This method delegates the serialization to the underlying
        `OfficeSettings` body, returning a dictionary representation of
        the settings.

        Returns:
            dict: A dictionary representing the settings content.
        """
        body: OfficeSettings = cast(OfficeSettings, self.body)
        return body.as_dict()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>

from collections.abc import Iterable

import pytest

from odfdo.document import Document
from odfdo.style import Style


@pytest.fixture
def document(samples) -> Iterable[Document]:
    yield Document(samples("example.odt"))


def test_insert_common_style(document):
    style = Style("paragraph", "MyStyle")
    document.insert_style(style)
    inserted_style = document.get_style("paragraph", "MyStyle")

    assert style.serialize() == inserted_style.serialize()


def test_insert_default_style(document):
    style = Style("paragraph", "MyStyle")
    document.insert_style(style, default=True)
    inserted_style = document.get_style("paragraph")
    expected = '<style:default-style style:family="paragraph"/>'
    assert inserted_style.serialize() == expected


def test_insert_automatic_style(document):
    style = Style("paragraph")
    document.insert_style(style, automatic=True)
    assert style.name is not None


def test_insert_style_name_as_argument(document):
    style = Style("paragraph")
    arg_name = "some_name"
    returned_name = document.insert_style(style, name=arg_name, automatic=True)
    assert style.name is not None
    assert returned_name == arg_name


def test_insert_style_from_string(document):
    style_str = (
        '<style:style style:name="style_as_str" '
        'style:family="paragraph">'
        "<style:text-properties "
        'fo:background-color="#ff0000" '
        'fo:color="#0000ff"/>'
        "</style:style>"
    )
    returned_name = document.insert_style(style_str, automatic=True)
    assert returned_name == "style_as_str"
    read_style = document.get_style("paragraph", "style_as_str")
    assert read_style.serialize() in (
        (
            '<style:style style:name="style_as_str" '
            'style:family="paragraph">'
            "<style:text-properties "
            'fo:background-color="#ff0000" '
            'fo:color="#0000ff"/>'
            "</style:style>"
        ),
        (
            '<style:style style:name="style_as_str" '
            'style:family="paragraph">'
            '<style:text-properties fo:color="#0000ff" '
            'fo:background-color="#ff0000"/>'
            "</style:style>"
        ),
        (
            '<style:style style:family="paragraph" '
            'style:name="style_as_str">'
            "<style:text-properties "
            'fo:background-color="#ff0000" '
            'fo:color="#0000ff"/>'
            "</style:style>"
        ),
    )


def test_insert_with_error(document):
    style = Style("paragraph", "MyStyle")
    with pytest.raises(AttributeError):
        document.insert_style(
            style=style,
            automatic=True,
            default=True,
        )


def test_insert_master_page_style(document):
    style = Style("master-page", "MyPageStyle")
    document.insert_style(style)
    inserted_style = document.get_style("master-page", "MyPageStyle")
    assert style.serialize() == inserted_style.serialize()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>
"""Section class for "text:section" tag."""

from __future__ import annotations

from typing import Any, Union, cast

from .element import Element, PropDef, register_element_class
from .mixin_link import LinkMixin
from .mixin_list import ListMixin
from .mixin_toc import TocMixin


class SectionMixin(Element):
    """Mixin class for elements that can contain sections.

    This mixin provides methods to access and manipulate "text:section"
    elements within a parent element.

    Used by the following classes:
        - "draw:text-box"
        - "office:text"
        - "style:footer"
        - "style:footer-left"
        - "style:header"
        - "style:header-left"
        - "table:covered-table-cell"
        - "table:table-cell"
        - "text:deletion"
        - "text:index-body"
        - "text:index-title"
        - "text:note-body"
        - "text:section"
    """

    def get_sections(
        self,
        style: str | None = None,
        content: str | None = None,
    ) -> list[Section]:
        """Return all sections that match the criteria.

        Args:
            style: The style name to filter by.
            content: A regex to match in the section content.

        Returns:
            list[Section]: A list of matching Section elements.
        """
        return cast(
            list[Section],
            self._filtered_elements("text:section", text_style=style, content=content),
        )

    @property
    def sections(
        self,
    ) -> list[Section]:
        """Return all sections within the element.

        Returns:
            list[Section]: A list of all Section elements.
        """
        return cast(list[Section], self.get_elements("text:section"))

    def get_section(
        self,
        position: int = 0,
        content: str | None = None,
    ) -> Section | None:
        """Return the section that matches the criteria.

        Args:
            position: The index of the section to return.
                Defaults to 0 (first section).
            content: A regex to match in the section content.

        Returns:
            Section or None: The matching Section element, or None if not found.
        """
        return cast(
            Union[None, Section],
            self._filtered_element(
                "descendant::text:section", position, content=content
            ),
        )


class Section(ListMixin, TocMixin, LinkMixin, SectionMixin):
    """Represents a section within a text document, "text:section".

    A section is a container for document content, often used for
    organization or to apply specific formatting or attributes.
    """

    _tag = "text:section"
    _properties = (
        PropDef("style", "text:style-name"),
        PropDef("name", "text:name"),
    )

    def __init__(
        self,
        style: str | None = None,
        name: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a Section element.

        Args:
            style: The style name for the section.
            name: The name of the section.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if style:
                self.style = style
            if name:
                self.name = name

    def get_formatted_text(self, context: dict | None = None) -> str:
        """Return the formatted text content of the section.

        Args:
            context: A dictionary of context variables.

        Returns:
            str: The formatted text content.
        """
        result = [element.get_formatted_text(context) for element in self.children]
        result.append("\n")
        return "".join(result)


Section._define_attribut_property()

register_element_class(Section)



#!/usr/bin/env python
"""Change an image in many ODF files.

This recipe is suitable for the scenario where an organization
is moving from one company logo to another and needs to replace
the logo in several hundred existing documents.
"""

import os
from hashlib import sha256
from pathlib import Path

from odfdo import Document

_DOC_SEQUENCE = 270
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "new_logo"
DATA = Path(__file__).parent / "data"
OLD_PRESENTATIONS = DATA / "old_presentations"
OLD_LOGO = OLD_PRESENTATIONS / "oldlogo.png"
NEW_LOGO = DATA / "newlogo.png"


def save_modified(document: Document) -> None:
    """Save a modified Document."""
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    name = Path(document.path).name
    new_path = OUTPUT_DIR / name
    print("Saving:", new_path)
    document.save(new_path)


def footprint(content: bytes) -> str:
    """Return sha256 digest of a bytes content."""
    return sha256(content).hexdigest()


def update_document_logo(
    path: Path,
    old_hash: str,
    new_content: bytes,
    stats: dict[str, int],
) -> None:
    """Update the logo inside a Document."""
    stats["files"] += 1
    if not path.suffix.lower().startswith(".od"):
        return
    try:
        document = Document(path)
    except Exception:
        return

    stats["odf_files"] += 1
    document_changed = False
    for image in document.body.images:
        image_url = image.url
        if not image_url:
            continue
        try:
            image_content = document.get_part(image_url)
        except KeyError:
            print("- not found inside document:", path, end=" ")
            print("  image URL:", image_url)
            continue
        if footprint(image_content) == old_hash:
            document.set_part(image_url, new_content)
            document_changed = True
    if document_changed:
        save_modified(document)
        stats["updated_files"] += 1


def update_logos() -> dict[str, int]:
    """Update logo image in all documents."""
    result: dict[str, int] = {
        "files": 0,
        "odf_files": 0,
        "updated_files": 0,
    }
    old_hash = footprint(OLD_LOGO.read_bytes())

    # making the new image content :
    buffer = Document("text")
    url = buffer.add_file(str(NEW_LOGO))
    new_content = buffer.get_part(url)

    for path in OLD_PRESENTATIONS.glob("**/*"):
        update_document_logo(path, old_hash, new_content, result)
    return result


def main() -> None:
    stats = update_logos()
    print(f"Files: {stats['files']}")
    print(f"ODF files: {stats['odf_files']}")
    print(f"Updated files: {stats['updated_files']}")
    test_unit(stats)


def test_unit(stats: dict[str, int]) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    assert (stats["files"]) == 3
    assert (stats["odf_files"]) == 2
    assert (stats["updated_files"]) == 2


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from collections.abc import Iterable

import pytest

from odfdo.const import ODF_STYLES
from odfdo.document import Document
from odfdo.style import Style
from odfdo.style_utils import (
    _check_background_support,
    _check_opacity,
    _check_position,
    _check_repeat,
    _erase_background,
    _map_key,
    _set_background,
    _set_background_color,
    _set_background_image,
)
from odfdo.styles import Styles


@pytest.fixture
def styles(samples) -> Iterable[Styles]:
    document = Document(samples("example.odt"))
    yield document.get_part(ODF_STYLES)


def test_check_background_support():
    assert _check_background_support("text") is None


def test_check_background_support_raise():
    with pytest.raises(TypeError):
        _check_background_support("oops")


def test_check_position_none():
    assert _check_position(None) is None


def test_check_position():
    assert _check_position("center") is None


def test_check_position_raise_1():
    with pytest.raises(ValueError):
        _check_position(" ")


def test_check_position_raise_2():
    with pytest.raises(ValueError):
        _check_position("oops")


def test_check_repeat_none():
    assert _check_repeat(None) is None


def test_check_repeat():
    assert _check_repeat("repeat") is None


def test_check_repeat_raise_1():
    with pytest.raises(ValueError):
        _check_repeat(" ")


def test_check_repeat_raise_2():
    with pytest.raises(ValueError):
        _check_repeat("oops")


def test_check_opacity_none():
    assert _check_opacity(None) is None


def test_check_opacity_int():
    assert _check_opacity(50) is None


def test_check_opacity_str():
    assert _check_opacity("50") is None


def test_check_opacity_raise_1():
    with pytest.raises(ValueError):
        _check_opacity("oops")


def test_check_opacity_raise_2():
    with pytest.raises(ValueError):
        _check_opacity(-1)


def test_check_opacity_raise_3():
    with pytest.raises(ValueError):
        _check_opacity(999)


def test_check_opacity_raise_4():
    with pytest.raises(ValueError):
        _check_opacity("-1")


def test_check_opacity_raise_5():
    with pytest.raises(ValueError):
        _check_opacity("999")


def test_find_page_layout(styles):
    auto = styles.office_automatic_styles
    layouts = auto.get_elements("style:page-layout")
    lay = layouts[0]
    assert lay.name == "Mpm1"


def test_erase_background(styles):
    auto = styles.office_automatic_styles
    layouts = auto.get_elements("style:page-layout")
    lay = layouts[0]
    assert _erase_background(lay) is None


def test_erase_background_2(styles):
    auto = styles.office_automatic_styles
    layouts = auto.get_elements("style:page-layout")
    lay = layouts[0]
    properties = lay.get_element("style:page-layout-properties")
    properties.delete()
    assert _erase_background(lay) is None


def test_set_background_color_1(styles):
    auto = styles.office_automatic_styles
    layouts = auto.get_elements("style:page-layout")
    lay = layouts[0]
    assert _set_background_color(lay, "#FFFFFF") is None


def test_erase_background_3(styles):
    auto = styles.office_automatic_styles
    layouts = auto.get_elements("style:page-layout")
    lay = layouts[0]
    _set_background_color(lay, "#FFFFFF")
    assert _erase_background(lay) is None


def test_set_background_image_1(styles):
    auto = styles.office_automatic_styles
    layouts = auto.get_elements("style:page-layout")
    lay = layouts[0]
    assert (
        _set_background_image(lay, "http://example.com/aaa", None, None, None, None)
        is None
    )


def test_erase_background_4(styles):
    auto = styles.office_automatic_styles
    layouts = auto.get_elements("style:page-layout")
    lay = layouts[0]
    _set_background_image(lay, "http://example.com/aaa", None, None, None, None)
    assert _erase_background(lay) is None


def test_set_background_color_2(styles):
    auto = styles.office_automatic_styles
    layouts = auto.get_elements("style:page-layout")
    lay = layouts[0]
    _set_background_image(lay, "http://example.com/aaa", None, None, None, None)
    assert _set_background_color(lay, "#FFFFFF") is None


def test_set_background_image_2(styles):
    auto = styles.office_automatic_styles
    layouts = auto.get_elements("style:page-layout")
    lay = layouts[0]
    _set_background_image(lay, "http://example.com/aaa", None, None, None, None)
    assert (
        _set_background_image(lay, "http://example.com/bbb", None, None, None, None)
        is None
    )


def test_set_background_1(styles):
    auto = styles.office_automatic_styles
    layouts = auto.get_elements("style:page-layout")
    lay = layouts[0]
    _set_background_image(lay, "http://example.com/aaa", None, None, None, None)
    assert _set_background(lay, None, None, None, None, None, None) is None


def test_set_background_text_style(styles):
    style = Style("text", name="name", area="text")
    with pytest.raises(TypeError):
        _set_background(style, None, "http://example.com/aaa", None, None, None, None)


def test_map_key_known():
    key = "chart:angle-offset"
    odf_key = _map_key(key)
    assert odf_key == key


def test_map_key_mapped_col():
    key = "display"
    odf_key = _map_key(key)
    assert odf_key == "text:display"


def test_map_key_mapped_no_mapped_no_col():
    key = "db:apply-command"
    odf_key = _map_key(key)
    assert odf_key == key


def test_map_key_mapped_no_mapped_no_col_not_known():
    key = "ooops"
    odf_key = _map_key(key)
    assert odf_key is None


def test_map_key_mapped_no_mapped_no_col_known():
    key = "border-left"
    odf_key = _map_key(key)
    assert odf_key == "fo:border-left"



#!/usr/bin/env python
"""Basic Accessibility test: check, for every picture in a document, if
there is:

  - a title (svg_title),
  - a description (svg_description)

or, at least, some caption text.

See test file `planes.odt` file and the result of the script.
"""

# Expected result on stdout:
# The document displays 3 pictures:
#  - pictures with a title: 2
#  - pictures with a description: 1
#  - pictures with a caption: 0

# Image: 100000000000013B000000D345859F604DCE636A.jpg
#   Name: graphics2, Title: Spitfire, general view, Description:Green spitfire in a hall, view from left front., Caption:None
# Image: 100000000000013B000000D3F908DA0A939D2F4B.jpg
#   Name: graphics3, Title: Spitfire, detail, Description:None, Caption:None
# Image: 100000000000013B000000D375CEBFD6D7CB7CE9.jpg
#   Name: graphics1, Title: None, Description:None, Caption:None

import os
import sys
from pathlib import Path
from typing import Any

from odfdo import Document

_DOC_SEQUENCE = 200
DATA = Path(__file__).parent / "data"
SOURCE = "planes.odt"


def read_source_document() -> Document:
    """Return the source Document."""
    try:
        source = sys.argv[1]
    except IndexError:
        source = DATA / SOURCE
    return Document(source)


def accessibility_evaluator(document: Document) -> dict[str, Any]:
    """Count for each images: titles, caption,description."""
    result: dict[str, Any] = {
        "images": [],
        "titles": 0,
        "descriptions": 0,
        "captions": 0,
    }

    # We want the images of the document.
    body = document.body
    images = body.images

    for image in images:
        uri = image.url
        filename = uri.rpartition("/")[2]
        frame = image.parent
        name = frame.name
        title = frame.svg_title
        description = frame.svg_description
        link = frame.parent
        # this part requires some ODF know how:
        caption = None
        if link.tag == "draw:a":
            caption = link.get_attribute("office:name")

        result["images"].append(
            f"Image: {filename}\n"
            f"  Name: {name}, Title: {title}, "
            f"Description:{description}, Caption:{caption}"
        )
        if title:
            result["titles"] += 1
        if description:
            result["descriptions"] += 1
        if caption:
            result["captions"] += 1

    return result


def display_accessibilty(stats: dict[str, Any]) -> None:
    """Print the stats on stdout."""
    print(f"The document displays {len(stats['images'])} pictures:")
    print(f" - pictures with a title: {stats['titles']}")
    print(f" - pictures with a description: {stats['descriptions']}")
    print(f" - pictures with a caption: {stats['captions']}")
    print()
    for content in stats["images"]:
        print(content)


def main() -> None:
    document = read_source_document()
    stats = accessibility_evaluator(document)
    display_accessibilty(stats)
    test_unit(stats)


def test_unit(stats: dict[str, Any]) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    assert len(stats["images"]) == 3
    assert stats["titles"] == 2
    assert stats["descriptions"] == 1
    assert stats["captions"] == 0


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Mixin classes for Form controls.

(The main objective of the current minimal implementation of forms is to parse
the existing form contents in a document.)"""

from __future__ import annotations

from decimal import Decimal
from typing import ClassVar

from .element import Element


class FormDelayRepeatMixin(Element):
    """Mixin for the "form:delay-for-repeat" attribute."""

    @property
    def delay_for_repeat(self) -> str:
        return self._get_attribute_str_default("form:delay-for-repeat", "PT0.050S")

    @delay_for_repeat.setter
    def delay_for_repeat(self, delay_for_repeat: str | None) -> None:
        self._set_attribute_str_default(
            "form:delay-for-repeat", delay_for_repeat, "PT0.050S"
        )


class FormMaxLengthMixin(Element):
    """Mixin for the "form:max-length" attribute."""

    @property
    def max_length(self) -> int | None:
        return self.get_attribute_integer("form:max-length")

    @max_length.setter
    def max_length(self, max_length: int | None) -> None:
        if max_length is None:
            self.del_attribute("form:max-length")
        else:
            max_length = max(max_length, 0)
        self._set_attribute_str_default("form:max-length", str(max_length), "")


class FormAsDictMixin:
    """Mixin for the as_dict() method of Form Control classes."""

    def as_dict(self) -> dict[str, str | Decimal | int | None]:
        """Serialize the Form content as a Python dict.

        Returns:
            The Form content as a Python dict.
        """
        return {
            "tag": self.tag,  # type: ignore[attr-defined]
            "name": self.name,  # type: ignore[attr-defined]
            "xml_id": self.xml_id,  # type: ignore[attr-defined]
            "value": self.value,  # type: ignore[attr-defined]
            "current_value": self.current_value
            if hasattr(self, "current_value")
            else None,
            "str": str(self),
        }


class FormSourceListMixin(Element):
    """Mixin for the "form:list-source-type" attribute."""

    LIST_SOURCE_TYPE: ClassVar[set[str]] = {
        "table",
        "query",
        "sql",
        "sql-pass-through",
        "value-list",
        "table-fields",
    }

    @property
    def list_source_type(self) -> str | None:
        return self.get_attribute_string("form:list-source-type")

    @list_source_type.setter
    def list_source_type(self, list_source_type: str | None) -> None:
        if list_source_type is None:
            self.del_attribute("form:list-source-type")
            return
        if list_source_type not in self.LIST_SOURCE_TYPE:
            raise ValueError
        self.set_attribute("form:list-source-type", list_source_type)


class FormSizetMixin(Element):
    """Mixin for the "form:size" attribute."""

    @property
    def size(self) -> int | None:
        return self.get_attribute_integer("form:size")

    @size.setter
    def size(self, size: int | None) -> None:
        if size is None:
            self.del_attribute("form:size")
        else:
            size = max(size, 0)
        self._set_attribute_str_default("form:size", str(size), "")


class FormImagePositionMixin(Element):
    """Mixin for the "form:image-position" attribute."""

    IMAGE_POSITION: ClassVar[set[str]] = {"bottom", "center", "end", "start", "top"}

    @property
    def image_position(self) -> str | None:
        return self._get_attribute_str_default("form:image-position", "center")

    @image_position.setter
    def image_position(self, image_position: str | None) -> None:
        if image_position is None or image_position in self.IMAGE_POSITION:
            self._set_attribute_str_default(
                "form:image-position", image_position, "center"
            )
        else:
            raise ValueError


class FormImageAlignMixin(Element):
    """Mixin for the "form:image-align" attribute."""

    IMAGE_ALIGN: ClassVar[set[str]] = {"start", "center", "end"}

    @property
    def image_align(self) -> str | None:
        return self._get_attribute_str_default("form:image-align", "center")

    @image_align.setter
    def image_align(self, image_align: str | None) -> None:
        if image_align is None or image_align in self.IMAGE_ALIGN:
            self._set_attribute_str_default("form:image-align", image_align, "center")
        else:
            raise ValueError


class FormButtonTypeMixin(Element):
    """Mixin for the "form:button-type" attribute."""

    BUTTON_TYPES: ClassVar[set[str]] = {"submit", "reset", "push", "url"}

    @property
    def button_type(self) -> str | None:
        return self._get_attribute_str_default("form:button-type", "push")

    @button_type.setter
    def button_type(self, button_type: str | None) -> None:
        if button_type is None or button_type in self.BUTTON_TYPES:
            self._set_attribute_str_default("form:button-type", button_type, "push")
        else:
            raise ValueError


class OfficeTargetFrameMixin(Element):
    """Mixin for the "office:target-frame" attribute.

    Usable with the following elements: "form:button", "form:form" and
    "form:imag".
    """

    TARGET_FRAME: ClassVar[set[str]] = {"_blank", "_parent", "_self", "_top"}

    @property
    def target_frame(self) -> str | None:
        return self._get_attribute_str_default("office:target-frame", "_blank")

    @target_frame.setter
    def target_frame(self, target_frame: str | None) -> None:
        # target_frame can be any str (frame name)
        self._set_attribute_str_default("office:target-frame", target_frame, "_blank")



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python


from odfdo.smil import AnimPar, AnimSeq, AnimTransFilter


def test_anim_par_class():
    ap = AnimPar()
    assert isinstance(ap, AnimPar)


def test_anim_seq_class():
    seq = AnimSeq()
    assert isinstance(seq, AnimSeq)


def test_anim_tf_class():
    atf = AnimTransFilter()
    assert isinstance(atf, AnimTransFilter)


def test_anim_par_node():
    ap = AnimPar(presentation_node_type="timing-root")
    assert ap.presentation_node_type == "timing-root"


def test_anim_par_smil_begin():
    ap = AnimPar(smil_begin="10s")
    assert ap.smil_begin == "10s"


def test_anim_seq_node():
    seq = AnimSeq(presentation_node_type="timing-root")
    assert seq.presentation_node_type == "timing-root"


def test_anim_tf_args():
    atf = AnimTransFilter(
        smil_dur="2s",
        smil_type="barWipe",
        smil_subtype="leftToRight",
        smil_direction="forward",
        smil_fadeColor="#000000",
        smil_mode="in",
    )
    assert atf.smil_dur == "2s"
    assert atf.smil_type == "barWipe"
    assert atf.smil_subtype == "leftToRight"
    assert atf.smil_direction == "forward"
    assert atf.smil_fadeColor == "#000000"
    assert atf.smil_mode == "in"



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>
#          Romain Gauthier <romain@itaapy.com>
"""List class for "text:list" tag and ListItem for "text:list-item" tag ."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Union, cast

from .element import (
    FIRST_CHILD,
    NEXT_SIBLING,
    PREV_SIBLING,
    Element,
    PropDef,
    register_element_class,
)
from .mixin_list import ListMixin
from .mixin_md import MDList, MDListItem
from .paragraph import Paragraph


class ListItem(MDListItem, ListMixin, Element):
    """An item of a list, "text:list-item"."""

    _tag = "text:list-item"

    def __init__(
        self,
        text_or_element: str | Element | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the ListItem, "text:list-item".

        A "ListItem" can be initialized with text content or another element.

        Args:
            text_or_element: The initial content of the list item. If a
                string, a paragraph containing the text is created. If an
                element, it is appended as a child.
            **kwargs: Additional keyword arguments for the parent `Element` class.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if isinstance(text_or_element, str):
                self.text_content = text_or_element
            elif isinstance(text_or_element, Element):
                self.append(text_or_element)
            elif text_or_element is not None:
                raise TypeError(f"Expected str or Element, not {type(text_or_element)}")

    def __str__(self) -> str:
        self._md_initialize_level()
        return "\n".join(self._md_collect())


class ListHeader(ListMixin, Element):
    """An header of a list, "text:list-header"."""

    _tag = "text:list-header"

    def __init__(
        self,
        text_or_element: str | Element | Iterable[str | Element] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the ListHeader, "text:list-header".

        A "ListItem" can be initialized with text content or another element.

        Args:
            text_or_element: The initial content of the list item. If a
                string, a paragraph containing the text is created. If an
                element, it is appended as a child.
            **kwargs: Additional keyword arguments for the parent `Element` class.
        """
        super().__init__(**kwargs)
        if self._do_init:
            actual_list: list[str | Element] | tuple = []
            if text_or_element is None:
                pass
            elif isinstance(text_or_element, (str, Element)):
                actual_list = [text_or_element]
            elif isinstance(text_or_element, (list, tuple)):
                actual_list = text_or_element
            else:
                raise TypeError(f"Expected str or Element, not {text_or_element!r}")
            for item in reversed(actual_list):
                if isinstance(item, str):
                    paragraph = Paragraph(item)
                else:
                    paragraph = item
                self.insert(paragraph, FIRST_CHILD)


class List(MDList, Element):
    """A list of elements, "text:list"."""

    _tag = "text:list"
    _properties = (PropDef("style", "text:style-name"),)

    def __init__(
        self,
        list_content: str | Element | Iterable[str | Element] | None = None,
        style: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the List.

        A `List` can be initialized with content, which can be a single item
        (string or element) or an iterable of items.

        Args:
            list_content: The initial content of the list. Each item is wrapped in a
                `ListItem`.
            style: The name of the style to apply to the list.
            **kwargs: Additional keyword arguments for the parent `Element` class.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if list_content:
                if isinstance(list_content, (Element, str)):
                    self.append(ListItem(list_content))
                elif hasattr(list_content, "__iter__"):
                    for item in list_content:
                        self.append(ListItem(item))
            if style is not None:
                self.style = style

    def get_items(self, content: str | None = None) -> list[Element]:
        """Get all list items (`ListItem`) within the list.

        Optionally filters items by their textual content.

        Args:
            content: A regular expression to match against
                the text content of the items.

        Returns:
            list[Element]: A list of `ListItem` elements that match the criteria.
        """
        return self._filtered_elements("text:list-item", content=content)

    def get_item(
        self,
        position: int = 0,
        content: str | None = None,
    ) -> Element | None:
        """Get a single list item from the list.

        Can retrieve an item by its position or by matching its text content.
        In nested lists, it returns the `ListItem` that directly contains
        the matched content.

        Args:
            position: The index of the item to retrieve. Defaults to 0.
            content: A regular expression to match against
                the text content of the items. If provided, `position` is
                ignored.

        Returns:
            Element | None: The matching `ListItem` element, or `None` if not found.
        """
        # Custom implementation because of nested lists
        if content:
            # Don't search recursively but on the very own paragraph(s) of
            # each list item
            for paragraph in self.get_elements("descendant::text:p"):
                if paragraph.match(content):
                    return paragraph.get_element("parent::text:list-item")
            return None
        return self._filtered_element("text:list-item", position)

    @property
    def list_header(self) -> ListHeader | None:
        """Get or set the list header."""
        return cast(Union[None, ListHeader], self.get_element("text:list-header"))

    @list_header.setter
    def list_header(
        self, text_or_element: str | Element | Iterable[str | Element] | None = None
    ) -> None:
        current = cast(Union[None, ListHeader], self.get_element("text:list-header"))
        if current:
            current.delete()
        new_header = ListHeader(text_or_element)
        self.insert(new_header, FIRST_CHILD)

    def set_list_header(
        self,
        text_or_element: str | Element | Iterable[str | Element],
    ) -> None:
        """Set the header of the list.

        This method replaces any existing header paragraphs (`text:p`) with
        the provided content.

        You may consider using the property List.list_header.

        Args:
            text_or_element: The content for the list header. Can be a single
                string or element, or an iterable of strings and/or elements.
        """
        self.list_header = text_or_element

    def insert_item(
        self,
        item: ListItem | str | Element | None,
        position: int | None = None,
        before: Element | None = None,
        after: Element | None = None,
    ) -> None:
        """Insert a new item into the list.

        The item can be inserted at a specific `position`, or `before` or
        `after` an existing element.

        Args:
            item: The item to insert.
                If not a `ListItem`, it will be wrapped in one.
            position: The index at which to insert the item.
            before: An existing element to insert the item
                before.
            after: An existing element to insert the item
                after.

        Raises:
            ValueError: If no position (`position`, `before`, or `after`) is specified.
        """
        if not isinstance(item, ListItem):
            item = ListItem(item)
        if before is not None:
            before.insert(item, xmlposition=PREV_SIBLING)
        elif after is not None:
            after.insert(item, xmlposition=NEXT_SIBLING)
        elif position is not None:
            self.insert(item, position=position)
        else:
            raise ValueError("Position must be defined")

    def append_item(
        self,
        item: ListItem | str | Element | None,
    ) -> None:
        """Append a new item to the end of the list.

        Args:
            item: The item to append.
                If not a `ListItem`, it will be wrapped in one.
        """
        if not isinstance(item, ListItem):
            item = ListItem(item)
        self.append(item)

    def get_formatted_text(self, context: dict | None = None) -> str:
        """Return a formatted string representation of the list.

        Each list item is prefixed with "- " and indented.

        Args:
            context: A dictionary providing context for formatting. If
                `rst_mode` is True in the context, additional
                newlines are added for reStructuredText compatibility.

        Returns:
            str: The formatted text content of the list.
        """
        if context is None:
            context = {
                "document": None,
                "footnotes": [],
                "endnotes": [],
                "annotations": [],
                "rst_mode": False,
                "img_counter": 0,
                "images": [],
                "no_img_level": 0,
            }
        rst_mode = bool(context.get("rst_mode"))
        result = []
        if rst_mode:
            result.append("\n")
        list_header = self.list_header
        if list_header:
            text_header_buf = []
            for child in list_header.children:
                text = child.get_formatted_text(context)
                text_header_buf.append(text)
                text_sum = "".join(text_header_buf)
                text_sum = text_sum.strip("\n")
                # Indent the text
                text_sum = text_sum.replace("\n", "\n  ")
                text_sum = f"  {text_sum}\n"
                result.append(text_sum)
        for list_item in self.get_elements("text:list-item"):
            textbuf = []
            for child in list_item.children:
                text = child.get_formatted_text(context)
                tag = child.tag
                if tag == "text:h":
                    # A title in a list is a bug
                    return text
                if tag == "text:list" and not text.lstrip().startswith("-"):
                    # If the list didn't indent, don't either
                    # (inner title)
                    return text  # pragma: nocover
                textbuf.append(text)
            text_sum = "".join(textbuf)
            text_sum = text_sum.strip("\n")
            # Indent the text
            text_sum = text_sum.replace("\n", "\n  ")
            text_sum = f"- {text_sum}\n"
            result.append(text_sum)
        if rst_mode:
            result.append("\n")
        return "".join(result)


List._define_attribut_property()

register_element_class(List)
register_element_class(ListHeader)
register_element_class(ListItem)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>


from odfdo.line_break import LineBreak


def test_create_line_break():
    lb = LineBreak()
    expected = "<text:line-break/>"
    assert lb.serialize() == expected


def test_create_line_break_str():
    lb = LineBreak()
    assert str(lb) == "\n"


def test_line_break_getter_0():
    lb = LineBreak()
    expected = "<text:line-break/>"
    assert lb.serialize() == expected


def test_line_break_getter_1():
    lb = LineBreak()
    assert lb.text == "\n"


def test_line_break_setter():
    lb = LineBreak()
    lb.text = "x"
    assert lb.text == "\n"



#!/usr/bin/env python
"""Create a basic text document with a list."""

import os
from pathlib import Path

from odfdo import Document, List, ListItem

_DOC_SEQUENCE = 20
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "basic_list"
TARGET = "document.odt"


def save_new(document: Document, name: str) -> None:
    """Save a recipe result Document."""
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    new_path = OUTPUT_DIR / name
    print("Saving:", new_path)
    document.save(new_path, pretty=True)


def generate_document() -> Document:
    """Generate a basic text document with a list."""
    # Create the document
    document = Document("text")
    body = document.body

    # Adding List
    my_list = List(["Arthur", "Ford", "Trillian"])
    # The list accepts a Python list of strings and list items.

    # The list can be written even though we will modify it afterwards:
    body.append(my_list)

    # Adding more List Item to the list
    item = ListItem("Marvin")
    my_list.append_item(item)

    # it should contain:
    print(document.get_formatted_text())
    # - Arthur
    # - Ford
    # - Trillian
    # - Marvin

    return document


def main() -> None:
    document = generate_document()
    test_unit(document)
    save_new(document, TARGET)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    assert str(document).strip() == "- Arthur\n- Ford\n- Trillian\n- Marvin"


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>

from collections.abc import Iterable

import pytest

from odfdo.document import Document
from odfdo.element import Element
from odfdo.paragraph import Paragraph
from odfdo.spacer import Spacer


@pytest.fixture
def document(samples) -> Iterable[Document]:
    document = Document(samples("base_text.odt"))
    yield document


def test_read_spacer_empty():
    element = Element.from_tag("<text:s />")
    assert isinstance(element, Spacer)
    assert element.length == 1


def test_read_spacer_empty_str():
    element = Element.from_tag("<text:s />")
    assert str(element) == " "


def test_read_spacer_1():
    element = Element.from_tag('<text:s text:c="1" />')
    assert isinstance(element, Spacer)
    assert element.length == 1


def test_read_spacer_1_str():
    element = Element.from_tag('<text:s text:c="1" />')
    assert str(element) == " "


def test_read_spacer_2():
    element = Element.from_tag('<text:s text:c="2" />')
    assert isinstance(element, Spacer)
    assert element.length == 2


def test_read_spacer_2_str():
    element = Element.from_tag('<text:s text:c="2" />')
    assert str(element) == "  "


def test_create_space_1_base():
    sp1 = Spacer()
    expected = "<text:s/>"
    assert sp1.serialize() == expected


def test_create_space_1_number():
    sp1 = Spacer()
    assert sp1.number in (None, "1")


def test_create_space_1_length():
    sp1 = Spacer()
    assert sp1.length == 1


def test_create_space_1_length_mod_serialize():
    sp1 = Spacer()
    sp1.length = 3
    expected = '<text:s text:c="3"/>'
    assert sp1.serialize() == expected


def test_create_space_1_length_mod_number():
    sp1 = Spacer()
    sp1.length = 3
    assert sp1.number == "3"


def test_create_space_1_length_mod_length():
    sp1 = Spacer()
    sp1.length = 3
    assert sp1.length == 3


def test_create_space_5_base():
    sp1 = Spacer(5)
    expected = '<text:s text:c="5"/>'
    assert sp1.serialize() == expected


def test_create_space_5_number():
    sp1 = Spacer(5)
    assert sp1.number == "5"


def test_create_space_5_length():
    sp1 = Spacer(5)
    assert sp1.length == 5


def test_create_spaces():
    sp5 = Spacer(5)
    expected = '<text:s text:c="5"/>'
    assert sp5.serialize() == expected


def test_create_naive_spaces():
    para = Paragraph("   ")
    # old rules
    # expected = '<text:p> <text:s text:c="2"/></text:p>'
    expected = '<text:p><text:s text:c="3"/></text:p>'
    assert para.serialize() == expected


def test_create_naive_spaces_no_format():
    para = Paragraph("   ", formatted=False)
    expected = "<text:p><text:s/></text:p>"
    assert para.serialize() == expected


def test_create_naive_cr_no_format():
    para = Paragraph("\n", formatted=False)
    expected = "<text:p><text:s/></text:p>"
    assert para.serialize() == expected


def test_plain_text_splitted_elements_spaces():
    txt = "A test,\n   \twith \n\n some é and \t and     5 spaces."
    para = Paragraph()
    content = para._expand_spaces(txt)
    content = para._merge_spaces(content)
    content = para._replace_tabs_lb(content)
    assert isinstance(content[3], Spacer)


def test_paragraph_1_space():
    para = Paragraph(" ")
    # new rule
    expected = "<text:p><text:s/></text:p>"
    assert para.serialize() == expected


def test_paragraph_2_spaces():
    para = Paragraph("  ")
    # new rule
    expected = '<text:p><text:s text:c="2"/></text:p>'
    assert para.serialize() == expected


def test_paragraph_1_space_no_format():
    para = Paragraph(" ", formatted=False)
    expected = "<text:p><text:s/></text:p>"
    assert para.serialize() == expected


def test_paragraph_2_spaces_no_format():
    para = Paragraph("  ", formatted=False)
    # new rule
    expected = "<text:p><text:s/></text:p>"
    assert para.serialize() == expected


def test_paragraph_any_spaces_no_format():
    para = Paragraph("  \n\t ", formatted=False)
    # new rule
    expected = "<text:p><text:s/></text:p>"
    assert para.serialize() == expected


def test_space_plus_space():
    para = Paragraph(" ")
    para.append(" ")
    expected = '<text:p><text:s text:c="2"/></text:p>'
    assert para.serialize() == expected


def test_space_plus_space_no_format():
    para = Paragraph(" ")
    para.append(" ", formatted=False)
    expected = '<text:p><text:s text:c="2"/></text:p>'
    assert para.serialize() == expected


def test_spacer_setter_0():
    spacer = Spacer()
    spacer.text = None
    expected = "<text:s/>"
    assert spacer.serialize() == expected


def test_spacer_setter_1():
    spacer = Spacer()
    spacer.text = ""
    expected = "<text:s/>"
    assert spacer.serialize() == expected


def test_spacer_setter_2():
    spacer = Spacer()
    spacer.text = "x"
    expected = "<text:s/>"
    assert spacer.serialize() == expected


def test_spacer_setter_3():
    spacer = Spacer()
    spacer.text = "xx"
    expected = '<text:s text:c="2"/>'
    assert spacer.serialize() == expected


def test_spacer_setter_4():
    spacer = Spacer()
    spacer.text = "xxx"
    expected = '<text:s text:c="3"/>'
    assert spacer.serialize() == expected



#!/usr/bin/env python
"""Create a presentation from a some images in a given directory,
where each image is put on the center of its own page scaled to either
the maximum available size, prefered maximum size, or cover the full
page and lose some info.
"""

import os
from pathlib import Path

# analyzing embedded image need Pillow library
from PIL import Image

from odfdo import Document, DrawPage, Frame

_DOC_SEQUENCE = 286
IMAGES = Path(__file__).parent / "data" / "images"
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "presentation_from_images"
TARGET = "presentation.odp"

MAX_SIZE = 15.0  # feel free to customize
CROP_SIZE = False  # feel free to customize

# Size (in cm) of a slide : (default page-layout)
SLIDE_W, SLIDE_H = 28.0, 21.0  # 4/3 screen

# FIXME: this is the default page-layout.
# - Changing the style of the page-layout by program is not done in this script
# - an other way, merging with external page-layout/master-page requires
#   extra files, out of the scope for this script.


def save_new(document: Document, name: str) -> None:
    """Save a recipe result Document."""
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    new_path = OUTPUT_DIR / name
    print("Saving:", new_path)
    document.save(new_path, pretty=True)


class ImageInfo:
    """Class to store informations about an image.

    Principle :
    - original image are left unmodified by the script
    - only the size they should appear is computed
    - later, the display engine (say LibreOffice) will merge this display
      information with other informations, like the size of the page
      (page-layout) and should act like a mask against the "big" croped image.
    """

    def __init__(self, path: Path) -> None:
        self.path = path
        self.size = None
        self.disp_w = self.disp_h = None
        self.pos_x = self.pos_y = None

    def adapt(self) -> bool:
        if not self.path.is_file():
            return False
        try:
            self.size = Image.open(self.path).size
        except OSError:
            # Not an image ?
            return False
        width, height = self.size
        if MAX_SIZE:
            ratio = max(width / MAX_SIZE, height / MAX_SIZE)
            display_w = width / ratio
            display_h = height / ratio
        elif CROP_SIZE:
            ratio = min(width / SLIDE_W, height / SLIDE_H)
            display_w = width / ratio
            display_h = height / ratio
        else:
            ratio = max(width / SLIDE_W, height / SLIDE_H)
            display_w = width / ratio
            display_h = height / ratio
        self.disp_w = f"{display_w:2f}cm"
        self.disp_h = f"{display_h:2f}cm"
        self.pos_x = f"{(SLIDE_W - display_w) / 2:2f}cm"
        self.pos_y = f"{(SLIDE_H - display_h) / 2:2f}cm"
        print(self.path.name, self.disp_w, self.disp_h)
        return True


def collect_images() -> list[ImageInfo]:
    pool = []
    for path in IMAGES.glob("**/*"):
        image_info = ImageInfo(path)
        if image_info.adapt():
            pool.append(image_info)
    return pool


def make_presentation(images_pool: list[ImageInfo]) -> Document:
    """Return a presentation made of images."""
    if not images_pool:  # unable to find images
        print("No image found !")
        return None

    presentation = Document("presentation")

    # Presentation got a body in which content is stored
    body = presentation.body
    body.clear()

    # For each image, we create a page in the presentation and display the image
    # and some text on this frame
    for image in images_pool:
        # add the file to the document
        uri = presentation.add_file(image.path)

        # Create an underlying page for the image and the text
        page = DrawPage(f"Page {image.path.name}")

        # Create a frame for the image
        image_frame = Frame.image_frame(
            image=uri,
            name=image.path.name,
            text="",  # Text over the image object
            size=(image.disp_w, image.disp_h),  # Display size of image
            anchor_type="page",
            page_number=None,
            position=(image.pos_x, image.pos_y),
            style=None,
        )

        page.append(image_frame)
        body.append(page)

    return presentation


def main() -> None:
    images_pool = collect_images()
    presentation = make_presentation(images_pool)
    if presentation is None:
        print("Something went wrong.")
        exit(0)
    test_unit(presentation)
    save_new(presentation, TARGET)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    body = document.body
    count = len([item for item in body.children if isinstance(item, DrawPage)])
    assert count == 3


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

import pytest

from odfdo.utils import hexa_color


def test_color_nothing():
    assert hexa_color() is None


def test_color_none():
    assert hexa_color(None) is None


def test_color_tuple():
    color = (171, 205, 239)
    expected = "#ABCDEF"
    assert hexa_color(color) == expected


def test_color_list():
    color = [171, 205, 239]
    with pytest.raises(TypeError):
        hexa_color(color)


def test_color_name():
    color = "violet"
    expected = "#EE82EE"
    assert hexa_color(color) == expected


def test_color_bad_name():
    color = "dark white"
    with pytest.raises(KeyError):
        hexa_color(color)


def test_color_bad_tuple():
    # For alpha channel? ;-)
    color = (171, 205, 238, 128)
    with pytest.raises(ValueError):
        hexa_color(color)


def test_color_bad_low_channel():
    color = (171, 205, -1)
    with pytest.raises(ValueError):
        hexa_color(color)


def test_color_bad_high_channel():
    color = (171, 205, 256)
    with pytest.raises(ValueError):
        hexa_color(color)


def test_color_bad_value():
    color = {}
    with pytest.raises(TypeError):
        hexa_color(color)


def test_color_empty_black():
    color = ""
    expected = "#000000"
    assert hexa_color(color) == expected


def test_color_already():
    color = "#000000"
    expected = "#000000"
    assert hexa_color(color) == expected


def test_color_already2():
    color = "#123456"
    expected = "#123456"
    assert hexa_color(color) == expected



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Command-line script to show or set user-defined fields in an ODF document.

This script allows for inspecting the values of user-defined fields and
updating them.
"""

from __future__ import annotations

import sys
from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from typing import TYPE_CHECKING

from odfdo import Document, Element, __version__
from odfdo.utils.script_utils import read_document, save_document

if TYPE_CHECKING:
    from odfdo.user_field_declaration import UserFieldDecl

PROG = "odfdo-userfield"


def configure_parser() -> ArgumentParser:
    description = "Inspect and modify user-defined fields within an ODF document."
    epilog = (
        "This tool allows you to view the values of specific user fields, "
        "list all available user fields, and update their values. "
        "User fields are custom data points stored in the document's metadata.\n\n"
        "Examples:\n"
        "  # Show the value of a specific user field named 'city'\n"
        "  $ odfdo-userfield -i file.odt -f city\n\n"
        "  # Show name, type, and value for fields 'city' and 'phone'\n"
        "  $ odfdo-userfield -i file.odt -ntf city phone\n\n"
        "  # Show all user fields with their name and value (repr format)\n"
        "  $ odfdo-userfield -i file.odt -anr\n\n"
        "  # Set the value of 'city' to 'Lyon' and 'phone' to '99 99', saving to result.odt\n"
        "  $ odfdo-userfield -i file.odt -o result.odt -s city Lyon -s phone '99 99'"
    )

    parser = ArgumentParser(
        prog=PROG,
        description=description,
        epilog=epilog,
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"{PROG} v{__version__}",
    )
    parser.add_argument(
        "-i",
        "--input",
        action="store",
        dest="input_file",
        metavar="INPUT",
        required=False,
        help="input file. if option not present, read from stdin",
    )
    parser.add_argument(
        "-o",
        "--output",
        action="store",
        dest="output_file",
        metavar="OUTPUT",
        required=False,
        help="If setting values, output file. if option not present, write to stdout",
    )
    parser.add_argument(
        "-f",
        "--field",
        action="extend",
        dest="fields",
        required=False,
        nargs=1,
        help="name of field to show",
    )
    parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        required=False,
        help="show all fields",
    )
    parser.add_argument(
        "-n",
        "--name",
        action="store_true",
        required=False,
        help="show field name",
    )
    parser.add_argument(
        "-t",
        "--type",
        action="store_true",
        required=False,
        help="show field type",
    )
    parser.add_argument(
        "-r",
        "--repr",
        action="store_true",
        required=False,
        help="format value using repr()",
    )
    parser.add_argument(
        "-s",
        "--set",
        action="append",
        dest="changes",
        required=False,
        nargs=2,
        help="set field value",
    )
    return parser


def parse_cli_args(cli_args: list[str] | None = None) -> Namespace:
    parser = configure_parser()
    return parser.parse_args(cli_args)


def main() -> None:
    args: Namespace = parse_cli_args()
    main_userfields(args)


def main_userfields(args: Namespace) -> None:
    try:
        document_userfields(args)
    except Exception as e:
        configure_parser().print_help()
        print()
        print(f"Error: {e.__class__.__name__}, {e}")
        raise SystemExit(1) from None


def document_userfields(args: Namespace) -> None:
    if args.all or args.fields:
        document = read_document(args.input_file)
        return show_fields(document, args)
    if args.changes:
        document = read_document(args.input_file)
        change_fields(document, args.changes)
        save_document(document, args.output_file)
    else:
        raise ValueError("missing arguments")


def _field_string(field: UserFieldDecl, args: Namespace) -> str:
    value, tpe = field.get_value(get_type=True)
    if args.repr:
        repr_value = repr(value)
    else:
        repr_value = str(value)
    line = []
    if args.name:
        line.append(f"{field.name}")
    if args.type:
        line.append(f"({tpe})")
    if line:
        return " ".join(line) + f": {repr_value}"
    return repr_value


def show_fields(document: Document, args: Namespace) -> None:
    field_set = set()
    if not args.all and args.fields:
        field_set = set(args.fields)
    body = document.body
    if hasattr(body, "get_user_field_decl_list"):
        for field in body.get_user_field_decl_list():
            if not args.all and field.name not in field_set:
                continue
            print(_field_string(field, args))
    else:  # pragma: nocover
        print("Warning: Document can not have user fields", file=sys.stderr)


def _change_field(body: Element, name: str, value: str) -> None:
    field = body.get_user_field_decl(name)  # type: ignore[attr-defined]
    if not field:
        print(f"Warning: unknown user-field {name!r}", file=sys.stderr)
        return
    field.set_value(value)


def change_fields(document: Document, changes: list[list[str]]) -> None:
    body = document.body
    if hasattr(body, "get_user_field_decl"):
        for name_value in changes:
            _change_field(body, name_value[0], name_value[1])
    else:  # pragma: nocover
        print("Warning: Document can not have user fields", file=sys.stderr)


if __name__ == "__main__":
    main()
